//---------------------------------------------------------------------------
// AnisotropicDisperse.cpp
//---------------------------------------------------------------------------
#include "AnisotropicDisperse.h"
#include "SimManager.h"
#include "ParsingFunctions.h"
#include "Plot.h"
#include "ModelMath.h"
#include <math.h>

///////////////////////////////////////////////////////////////////////////////
//     Constructor
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)

///////////////////////////////////////////////////////////////////////////////
clAnisotropicDisperse::clAnisotropicDisperse( clSimManager * p_oSimManager ) : 
     clWorkerBase( p_oSimManager ), clBehaviorBase( p_oSimManager ), 
     clDisperseBase( p_oSimManager ), clGridBasedSpatialDisperse( p_oSimManager )
     {
       try
       {
         strcpy( m_cNameString, "anisotropic disperse" );


         mp_fDirectionMaxDispersal = NULL;
         mp_fAnisotropicAmplitude = NULL;
         mp_fNormalizer = NULL;
         mp_fProb = NULL;

       }
       catch ( modelErr & err )
       {
         throw( err );
       }
       catch ( modelMsg & msg )
       {
         throw( msg );
       } //non-fatal error
       catch ( ... )
       {
         modelErr stcErr;
         stcErr.iErrorCode = UNKNOWN;
         strcpy( stcErr.cFunction, "clAnisotropicDisperse::clAnisotropicDisperse" );
         throw( stcErr );
       }
}

//---------------------------------------------------------------------------

///////////////////////////////////////////////////////////////////////////////
//   Destructor
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)

///////////////////////////////////////////////////////////////////////////////
clAnisotropicDisperse::~clAnisotropicDisperse()
{
  short int i, j; //loop counter

  if ( mp_fProb )
  {
    for ( i = 0; i < m_iNumBehaviorSpecies; i++ )
    {
      for ( j = 0; j < m_iNumAngles; j++ )
      {
        delete[] mp_fProb[i] [j];
      }
      delete[] mp_fProb[i];
    }
  }

  delete[] mp_fProb;


  delete[] mp_fDirectionMaxDispersal;
  delete[] mp_fAnisotropicAmplitude;
  delete[] mp_fNormalizer;
}

//---------------------------------------------------------------------------

///////////////////////////////////////////////////////////////////////////////
//   CalculateNormalizer
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)

///////////////////////////////////////////////////////////////////////////////
float clAnisotropicDisperse::CalculateNormalizer( int iSpecies )
{
  try
  {
    float fMean, //term that contains the angle
         fAngle, //angle to a point
         fNormalizer = 0, //normalizer to be returned
         fX, fY, //coordinates for looping around a circle
         //fExp, //value of the exponential for a distance
         //fTemp, //helps in exponential calcs
         fDist; //distance to point
    int iMaxDist = 1000, //max distance diameter in meters - search radius is
         //actually half this
         iHalfMax = (int)floor( ( float )iMaxDist / 2 ), iNegHalfMax = -1 * iHalfMax, i, j;
    short int iSpIndex; //species index

    iSpIndex = mp_iIndexes[iSpecies];

    for ( i = iNegHalfMax; i < iHalfMax; i++ )
    {
      for ( j = iNegHalfMax; j < iHalfMax; j++ )
      {

        //Get the angle to the point of the loops
        fX = ( 0.5 * i ) + 0.25;
        fY = ( 0.5 * j ) + 0.25;
        if ( ( fX >= 0 ) && ( fY > 0 ) ) fAngle = atan( fX / fY );
        if ( ( fX > 0 ) && ( fY <= 0 ) ) fAngle = ( M_PI / 2.0 ) + atan( ( 0 - fY ) / fX );
        if ( ( fX <= 0 ) && ( fY < 0 ) ) fAngle = M_PI + atan( ( 0 - fX ) / ( 0 - fY ) );
        if ( ( fX < 0 ) && ( fY >= 0 ) ) fAngle = ( 1.5 * M_PI ) + atan( fY / ( 0 - fX ) );

        //Get the distance
        fDist = sqrt( pow( fX, 2 ) + pow( fY, 2 ) );

        //Calculate the angle-based term
        fMean = ( mp_fDispersalX0[iSpIndex]
             - ( mp_fAnisotropicAmplitude[iSpIndex] * cos( fAngle - mp_fDirectionMaxDispersal[iSpIndex] ) ) );

        if ( fMean <= 0 ) fMean = 0.000001;

        //What is the function?
        if ( lognormal == mp_iWhatFunction[iSpIndex] )
        {
          fNormalizer += clModelMath::CalculateLognormalFunction( fMean, mp_fThetaXb[iSpIndex], fDist );
        }
        else if ( weibull == mp_iWhatFunction[iSpIndex] )
        {
          fNormalizer += clModelMath::CalculateWeibullFunction( fMean, mp_fThetaXb[iSpIndex], fDist );
        }
      }
    }

    fNormalizer *= 0.25;
    return fNormalizer;
  }
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clAnisotropicDisperse::CalculateNormalizer" );
    throw( stcErr );
  }
}

//---------------------------------------------------------------------------

///////////////////////////////////////////////////////////////////////////////
//     DoChildSetup()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)

///////////////////////////////////////////////////////////////////////////////
void clAnisotropicDisperse::DoChildSetup( xercesc::DOMDocument * p_oDoc )
{
  try
  {
    GetParameterData( p_oDoc );
    CalculateProbs();
  }
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clAnisotropicDisperse::DoSetup" );
    throw( stcErr );
  }

}

//---------------------------------------------------------------------------

//////////////////////////////////////////////////////////////////////////////
//     GetParameterData
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)

///////////////////////////////////////////////////////////////////////////////
void clAnisotropicDisperse::GetParameterData( xercesc::DOMDocument * p_oDoc )
{

  clPopulationBase * p_oTemp = mp_oSimManager->GetPopulationObject( "treepopulation" );
  clTreePopulation * p_oPop = ( clTreePopulation * ) p_oTemp;
  DOMElement * p_oElement = p_oDoc->getDocumentElement();
  floatVal * p_fTemp; //for getting species-specific values
  short int i; //loop counter

  //Declare our arrays
  mp_fDirectionMaxDispersal = new float[m_iNumBehaviorSpecies];
  mp_fAnisotropicAmplitude = new float[m_iNumBehaviorSpecies];


  //Set up our temp float array and pre-load with this behavior's species
  p_fTemp = new floatVal[m_iNumBehaviorSpecies];
  for ( i = 0; i < m_iNumBehaviorSpecies; i++ )
    p_fTemp[i].code = mp_iWhatSpecies[i];


  //Direction max dispersal
  FillSpeciesSpecificValue( p_oElement, "di_directionMaxDispersal", "di_dmdVal", p_fTemp,
       m_iNumBehaviorSpecies, p_oPop, true );
  for ( i = 0; i < m_iNumBehaviorSpecies; i++ )
    mp_fDirectionMaxDispersal[mp_iIndexes[p_fTemp[i].code]] = p_fTemp[i].val;

  //Anisotropic amplitude
  FillSpeciesSpecificValue( p_oElement, "di_anisotropicAmplitude", "di_aaVal", p_fTemp, m_iNumBehaviorSpecies, p_oPop, true );
  for ( i = 0; i < m_iNumBehaviorSpecies; i++ )
    mp_fAnisotropicAmplitude[mp_iIndexes[p_fTemp[i].code]] = p_fTemp[i].val;


  delete[] p_fTemp;
}

//---------------------------------------------------------------------------

///////////////////////////////////////////////////////////////////////////////
//   CalculateProbs
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)

///////////////////////////////////////////////////////////////////////////////
void clAnisotropicDisperse::CalculateProbs()
{
  try
  {
    //float fNormalizer, fAngle, //the angle for which the exponential is being calculated
    //     fMean; //term including the angle
    int i; //j, k, iSpIndex;

    mp_fNormalizer = new float[m_iNumBehaviorSpecies];
    //Now create the values in the arrays for each species
    for ( i = 0; i < m_iNumBehaviorSpecies; i++ )
    {

      //Calculate the normalizer
      mp_fNormalizer[i] = CalculateNormalizer( mp_iWhatSpecies[i] );
    }

  }
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clIsotropicDisperse::CalculateProbs" );
    throw( stcErr );
  }
}

//---------------------------------------------------------------------------
/*
///////////////////////////////////////////////////////////////////////////////
//   CalculateProbs
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)

///////////////////////////////////////////////////////////////////////////////
void clAnisotropicDisperse::CalculateProbs()
{
  try
  {
    float fNormalizer, fAngle, //the angle for which the exponential is being calculated
         fMean; //term including the angle
    int i, j, k, iSpIndex;

    //Set the number of angles for the array - every 5 degrees
    m_iNumAngles = 72;

    //Angle increment, in radians
    m_fAngleIncr = 2 * M_PI / m_iNumAngles;

    //Determine what the max search radius as an integer is
    m_iMaxSearchRadius = ceil( m_fMaxSearchRadius );

    //Declare the probability arrays
    mp_fProb = new float * * [m_iNumBehaviorSpecies];
    for ( i = 0; i < m_iNumBehaviorSpecies; i++ )
    {
      mp_fProb[i] = new float * [m_iNumAngles];
      for ( j = 0; j < m_iNumAngles; j++ )
      {
        mp_fProb[i] [j] = new float[m_iMaxSearchRadius];
      }
    }

    //Now create the values in the arrays for each species
    for ( i = 0; i < m_iNumBehaviorSpecies; i++ )
    {

      iSpIndex = mp_iIndexes[mp_iWhatSpecies[i]];

      //Calculate the normalizer
      fNormalizer = CalculateNormalizer( mp_iWhatSpecies[i] );

      //Since we're dividing by it, take its inverse so we can multiply instead
      fNormalizer = 1.0 / fNormalizer;

      //Loop out the array and calculate the function distance divided by the
      //normalizer
      for ( j = 0; j < m_iNumAngles; j++ )
      {

        //Get the angle
        fAngle = m_fAngleIncr * j;

        for ( k = 0; k < m_iMaxSearchRadius; k++ )

          fMean = ( mp_fDispersalX0[iSpIndex]
                   - ( mp_fAnisotropicAmplitude[iSpIndex] * cos( fAngle - mp_fDirectionMaxDispersal[iSpIndex] ) ) );

          if ( fMean <= 0 ) fMean = 0.000001;

          if ( lognormal == mp_iWhatFunction[iSpIndex] )
          {

            //Normally there would be an area term - but I want an area of one
            //square meter so I'm leaving out that term
            mp_fProb[i] [j] [k] = fNormalizer * CalculateLognormalFunction( fMean, mp_fThetaXb[iSpIndex], k + 1 );
          }
          else
          {
            //Normally there would be an area term - but I want an area of one
            //square meter so I'm leaving out that term
            mp_fProb[i] [j] [k] = fNormalizer * CalculateWeibullFunction( fMean, mp_fThetaXb[iSpIndex], k + 1 );
          }
      }
    }
  }
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clIsotropicDisperse::CalculateProbs" );
    throw( stcErr );
  }
}

//---------------------------------------------------------------------------
*/

///////////////////////////////////////////////////////////////////////////////
//   GetFunctionValue
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)

///////////////////////////////////////////////////////////////////////////////
float clAnisotropicDisperse::GetFunctionValue( float & fCellX, float & fCellY, float & fTreeX, float & fTreeY, int & iSpecies )
{
  try
  {

    //Get the distance
    float fDistance = mp_oPlot->GetDistance( fCellX, fCellY, fTreeX, fTreeY );
    if ( fDistance <= m_fMaxSearchRadius)
    {
      int iSpIndex = mp_iIndexes[iSpecies];
      float fAngle = mp_oPlot->GetAzimuthAngle( fCellX, fCellY, fTreeX, fTreeY );
      float fMean = ( mp_fDispersalX0[iSpIndex]
                 - ( mp_fAnisotropicAmplitude[iSpIndex] * cos( fAngle - mp_fDirectionMaxDispersal[iSpIndex] ) ) );

      if ( fMean <= 0 ) fMean = 0.000001;

      if ( lognormal == mp_iWhatFunction[iSpIndex] )
      {

            //Normally there would be an area term - but I want an area of one
            //square meter so I'm leaving out that term
        float fFunction = clModelMath::CalculateLognormalFunction( fMean, mp_fThetaXb[iSpIndex], fDistance );
        return fFunction / mp_fNormalizer[iSpIndex];
      }
      else
      {
            //Normally there would be an area term - but I want an area of one
            //square meter so I'm leaving out that term
        float fFunction = clModelMath::CalculateWeibullFunction( fMean, mp_fThetaXb[iSpIndex], fDistance );
        return fFunction / mp_fNormalizer[iSpIndex];
      }
    }
    else
    {
      return 0.0;
    }
  }
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clAnsotropicDisperse::GetFunctionValue" );
    throw( stcErr );
  }
}
//---------------------------------------------------------------------------
/*
///////////////////////////////////////////////////////////////////////////////
//   GetFunctionValue
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)

///////////////////////////////////////////////////////////////////////////////
float clAnisotropicDisperse::GetFunctionValue( float & fCellX, float & fCellY, float & fTreeX, float & fTreeY, int & iSpecies )
{
  try
  {

    //Get the distance
    int iDistance = RandomRound( mp_oPlot->GetDistance( fCellX, fCellY, fTreeX, fTreeY ) );
    if ( iDistance < m_iMaxSearchRadius )
    {
      int iAngle = RandomRound( mp_oPlot->GetAzimuthAngle( fCellX, fCellY, fTreeX, fTreeY ) / m_fAngleIncr );
      if (iAngle >= m_iNumAngles) iAngle = m_iNumAngles - 1;
      return mp_fProb[mp_iIndexes[iSpecies]] [iAngle] [iDistance];
    }
    else
    {
      return 0.0;
    }
  }
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clAnsotropicDisperse::GetFunctionValue" );
    throw( stcErr );
  }
}
//---------------------------------------------------------------------------
*/
