//---------------------------------------------------------------------------
#include "BasalAreaLight.h"
#include "ParsingFunctions.h"
#include "SimManager.h"
#include "GridBase.h"
#include "TreePopulation.h"
#include "Plot.h"
#include "ModelMath.h"
#include <stdio.h>

////////////////////////////////////////////////////////////////////////////
// Constructor
// Edit history:
// -------------
// March 2, 2006 - Created (LEM)
////////////////////////////////////////////////////////////////////////////
clBasalAreaLight::clBasalAreaLight( clSimManager * p_oSimManager ) : clWorkerBase( p_oSimManager ), 
  clBehaviorBase( p_oSimManager ), clLightBase(p_oSimManager)
{

  //Set the namestring
  strcpy( m_cNameString, "basalarealightshell" );

  //Allowed file types
  m_iNumAllowedTypes = 2;
  mp_iAllowedFileTypes = new int[m_iNumAllowedTypes];
  mp_iAllowedFileTypes[0] = parfile;
  mp_iAllowedFileTypes[1] = detailed_output;

  //Versions
  m_fVersionNumber = 1;
  m_fMinimumVersionNumber = 1;

  m_bNeedsCommonParameters = false;
  
  mp_iSpeciesTypes = NULL;
}

////////////////////////////////////////////////////////////////////////////
// Destructor
// Edit history:
// -------------
// March 2, 2006 - Created (LEM)
////////////////////////////////////////////////////////////////////////////
clBasalAreaLight::~clBasalAreaLight()
{
  delete[] mp_iSpeciesTypes;
}

////////////////////////////////////////////////////////////////////////////
// DoShellSetup()
// Edit history:
// -------------
// March 2, 2006 - Created (LEM)
////////////////////////////////////////////////////////////////////////////
void clBasalAreaLight::DoShellSetup( xercesc::DOMDocument * p_oDoc )
{
  try
  {
  	clTreePopulation *p_oPop = (clTreePopulation*) mp_oSimManager->GetPopulationObject("treepopulation");
    clPlot *p_oPlot = mp_oSimManager->GetPlotObject();
    DOMElement * p_oElement = p_oDoc->getDocumentElement();
    float fLight, fConBA, fAngBA;
    int i, j, iNumXCells, iNumYCells,
        iNumSpecies = p_oPop->GetNumberOfSpecies();
    bool bMap;
    
    int *p_iTemp = new int[iNumSpecies];

    m_bGridUpdated = false;
    
    mp_iSpeciesTypes = new iTreeType[iNumSpecies];

    //**********************************************
    //Read values from the parameter file
    //**********************************************
    //GLI mean value "a" parameter
    FillSingleValue( p_oElement, "li_baLightA", & m_fA, true );

    //Conifer "b" parameter
    FillSingleValue( p_oElement, "li_baConiferLightB", & m_fConiferB, true );
    
    //Angiosperm "b" parameter
    FillSingleValue( p_oElement, "li_baAngiospermLightB", & m_fAngiospermB, true );

    //Conifer "c" parameter
    FillSingleValue( p_oElement, "li_baConiferLightC", & m_fConiferC, true );
    //Error if this is equal to 0
    if (0 == m_fConiferC) {
      modelErr stcErr;
      strcpy( stcErr.cFunction, "clBasalAreaLight::DoShellSetup" );
      strcpy( stcErr.cMoreInfo, "Basal area light conifer \"c\" parameter cannot equal 0." );
      stcErr.iErrorCode = BAD_DATA;
      throw( stcErr );
    }
    
    //Angiosperm "c" parameter
    FillSingleValue( p_oElement, "li_baAngiospermLightC", & m_fAngiospermC, true );
    //Error if this is equal to 0
    if (0 == m_fAngiospermC) {
      modelErr stcErr;
      strcpy( stcErr.cFunction, "clBasalAreaLight::DoShellSetup" );
      strcpy( stcErr.cMoreInfo, "Basal area light angiosperm \"c\" parameter cannot equal 0." );
      stcErr.iErrorCode = BAD_DATA;
      throw( stcErr );
    }

    //Sigma parameter for lognormal PDF
    FillSingleValue( p_oElement, "li_baLightSigma", & m_fSigma, true );

    //Basal area change threshold for grid cells for recalculating GLI
    FillSingleValue( p_oElement, "li_baLightChangeThreshold", & m_fChangeThreshold, true );
    //Error if this is less than 0
    if (m_fChangeThreshold < 0) {
      modelErr stcErr;
      strcpy( stcErr.cFunction, "clBasalAreaLight::DoShellSetup" );
      strcpy( stcErr.cMoreInfo, "Basal area light change threshold cannot be less than 0." );
      stcErr.iErrorCode = BAD_DATA;
      throw( stcErr );
    }

    //Density light minimum DBH (cm) for a tree counting towards the density
    FillSingleValue( p_oElement, "li_baLightMinDBH", & m_fMinDbh, true );
    //Error if this is equal to 0
    if (m_fMinDbh < 0) {
      modelErr stcErr;
      strcpy( stcErr.cFunction, "clBasalAreaLight::DoShellSetup" );
      strcpy( stcErr.cMoreInfo, "Basal area light minimum DBH cannot be less than 0." );
      stcErr.iErrorCode = BAD_DATA;
      throw( stcErr );
    }
    
    //Whether a species is an angiosperm or conifer
    FillSpeciesSpecificValue( p_oElement, "li_baTreeType", "li_bttVal", p_iTemp, p_oPop, true );
    for (i = 0; i < iNumSpecies; i++) {
      if (angiosperm == p_iTemp[i]) mp_iSpeciesTypes[i] = angiosperm;
      else if (conifer == p_iTemp[i]) mp_iSpeciesTypes[i] = conifer;
      else {
      	modelErr stcErr;
        strcpy( stcErr.cFunction, "clBasalAreaLight::DoShellSetup" );
        sprintf( stcErr.cMoreInfo, "%s%d", "Unrecognized basal area light tree type: ", p_iTemp[i] );
        stcErr.iErrorCode = BAD_DATA;
        throw( stcErr );
      }
    }

    //**********************************************
    //Set up the light grid
    //**********************************************
    //Is there already a grid from the parameter file?
    mp_oLightGrid = mp_oSimManager->GetGridObject( "Basal Area Light" );

    if ( !mp_oLightGrid )
    {
      //Create the grid with three float data members
      mp_oLightGrid = mp_oSimManager->CreateGrid( "Basal Area Light", 0, 3, 0, 0 );
      //Register the data member called "Light"
      m_iGridLightCode = mp_oLightGrid->RegisterFloat( "Light" );
      //Register the data member called "Con BA"
      m_iGridConBACode = mp_oLightGrid->RegisterFloat( "Con BA" );
      //Register the data member called "Ang BA"
      m_iGridAngBACode = mp_oLightGrid->RegisterFloat( "Ang BA" );
    }
    else
    {
      //Get the data member codes
      m_iGridLightCode = mp_oLightGrid->GetFloatDataCode( "Light" );
      m_iGridConBACode = mp_oLightGrid->GetFloatDataCode( "Con BA" );
      m_iGridAngBACode = mp_oLightGrid->GetFloatDataCode( "Ang BA" );
      if ( -1 == m_iGridLightCode || -1 == m_iGridConBACode || -1 == m_iGridAngBACode)
      {
        modelErr stcErr;
        stcErr.iErrorCode = BAD_DATA;
        strcpy( stcErr.cFunction, "clBasalAreaLight::DoShellSetup" );
        strcpy( stcErr.cMoreInfo,
             "\"Basal Light\" grid was incorrectly set up in the parameter file.  Data members missing." );
        throw( stcErr );
      }
    }

    //Verify that the grid cell lengths divide evenly into the plot lengths
    iNumXCells = mp_oLightGrid->GetNumberXCells();
    iNumYCells = mp_oLightGrid->GetNumberYCells();
    if (fabs(p_oPlot->GetXPlotLength() - (iNumXCells * mp_oLightGrid->GetLengthXCells())) > 0.01 ||
        fabs(p_oPlot->GetYPlotLength() - (iNumYCells * mp_oLightGrid->GetLengthYCells())) > 0.01) {
      modelErr stcErr;
      stcErr.iErrorCode = BAD_DATA;
      strcpy( stcErr.cFunction, "clBasalAreaLight::DoShellSetup" );
      strcpy( stcErr.cMoreInfo,
                "\"Basal Area Light\" grid cell lengths must divide evenly into the plot lengths." );
      throw( stcErr );
    }

    //Set the basal area totals of trees for each grid cell to below the 
    //minimum change threshold, to be sure that a light level is calculated for 
    //each cell the first timestep; but first make sure that there's no grid map
    bMap = false;
    for (i = 0; i < iNumXCells; i++) {
      for (j = 0; j < iNumYCells; j++) {
        mp_oLightGrid->GetValueOfCell(i, j, m_iGridLightCode, &fLight);
        mp_oLightGrid->GetValueOfCell(i, j, m_iGridConBACode, &fConBA);
        mp_oLightGrid->GetValueOfCell(i, j, m_iGridAngBACode, &fAngBA);
        if (fLight > 0 || fConBA > 0 || fAngBA > 0) {
          bMap = true;
          break;
        }
      }
    }

    if (!bMap) {
      fConBA = -1 - m_fChangeThreshold;
      for (i = 0; i < iNumXCells; i++) {
        for (j = 0; j < iNumYCells; j++) {
          mp_oLightGrid->SetValueOfCell(i, j, m_iGridConBACode, fConBA);
        }
      }
    }
    
    delete[] p_iTemp;
  }
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clBasalAreaLight::DoShellSetup" );
    throw( stcErr );
  }
}

////////////////////////////////////////////////////////////////////////////
// CalcLightValue
// Edit history:
// -------------
// March 2, 2006 - Created (LEM)
////////////////////////////////////////////////////////////////////////////
float clBasalAreaLight::CalcLightValue(clTree *p_oTree, clTreePopulation *p_oPop)
{
  int iSp = p_oTree->GetSpecies(), iTp = p_oTree->GetType();
  float fX, fY, fGLI;

  //Check to see if the light grid has been updated this timestep, and do it
  //if it hasn't
  if (!m_bGridUpdated) UpdateGridValues(p_oPop);

  //Get the tree's location, then return the GLI from that cell of "Density
  //Light"
  p_oTree->GetValue(p_oPop->GetXCode(iSp, iTp), &fX);
  p_oTree->GetValue(p_oPop->GetYCode(iSp, iTp), &fY);
  mp_oLightGrid->GetValueAtPoint(fX, fY, m_iGridLightCode, &fGLI);
  return fGLI;
}

////////////////////////////////////////////////////////////////////////////
// UpdateGridValues()
// Edit history:
// -------------
// March 2, 2006 - Created (LEM)
////////////////////////////////////////////////////////////////////////////
void clBasalAreaLight::UpdateGridValues(clTreePopulation * p_oPop)
{
  try
  {
    clTreeSearch * p_oTrees; //saplings and adults
    clTree * p_oTree; //single tree to count
    char cQuery[50]; //format search string into this
    float **p_fConBAGrid, //holds conifer basal area for this timestep
          **p_fAngBAGrid; //holds angiosperm basal area for this timestep
    float fLight, //light level
          fDbh, //tree's DBH
          fBA, //tree's basal area
          fConBA, fAngBA, //conifer and angiosperm basal areas
          fX, fY, //holders for the tree's X and Y location
          fXCellLength = mp_oLightGrid->GetLengthXCells(),
          fYCellLength = mp_oLightGrid->GetLengthYCells();
    int iSp, iTp,
        iNumXCells = mp_oLightGrid->GetNumberXCells(),
        iNumYCells = mp_oLightGrid->GetNumberYCells(),
        iX, iY; //loop counter

    p_fConBAGrid = new float*[iNumXCells];
    p_fAngBAGrid = new float*[iNumXCells];
    for (iX = 0; iX < iNumXCells; iX++ ) {
      p_fConBAGrid[iX] = new float[iNumYCells];
      p_fAngBAGrid[iX] = new float[iNumYCells];
      for ( iY = 0; iY < iNumYCells; iY++ ) {
        p_fConBAGrid[iX][iY] = 0;
        p_fAngBAGrid[iX][iY] = 0;
      }
    }

    //Get all saplings and adults and count their basal area into our 
    //temporary grid
    sprintf(cQuery, "%s%i%s%i", "type=",clTreePopulation::sapling,",",clTreePopulation::adult);
    p_oTrees = p_oPop->Find(cQuery);
    p_oTree = p_oTrees->NextTree();
    while (NULL != p_oTree) {

      //Get the tree's DBH and make sure it's eligible
      iSp = p_oTree->GetSpecies();
      iTp = p_oTree->GetType();
      p_oTree->GetValue(p_oPop->GetDbhCode(iSp, iTp), &fDbh);

      if (fDbh >= m_fMinDbh) {
        //Get the tree's location
        p_oTree->GetValue(p_oPop->GetXCode(iSp, iTp), &fX);
        p_oTree->GetValue(p_oPop->GetYCode(iSp, iTp), &fY);
        
        //Get tree's basal area
        fBA = clModelMath::CalculateBasalArea(fDbh);
        
        //Increment the cell's basal area total
        if (angiosperm == mp_iSpeciesTypes[iSp])
          p_fAngBAGrid[(int)(fX / fXCellLength)][(int)(fY / fYCellLength)]+= fBA;
        else
          p_fConBAGrid[(int)(fX / fXCellLength)][(int)(fY / fYCellLength)]+= fBA;
      }
      p_oTree = p_oTrees->NextTree();
    }

    //Calculate each cell's light level
    for ( iX = 0; iX < iNumXCells; iX++ )
    {
      for ( iY = 0; iY < iNumYCells; iY++ )
      {
        //Has the total basal area changed by more than the minimum threshold?
        mp_oLightGrid->GetValueOfCell(iX, iY, m_iGridConBACode, &fConBA);
        mp_oLightGrid->GetValueOfCell(iX, iY, m_iGridAngBACode, &fAngBA);
        fBA = fConBA + fAngBA;
        if (fabs((p_fAngBAGrid[iX][iY] + p_fConBAGrid[iX][iY]) - fBA) > m_fChangeThreshold) {

          //Use the density to calculate a mean GLI
          fLight = GetMeanGLI(p_fConBAGrid[iX][iY], p_fAngBAGrid[iX][iY]);

          //Calculate the location parameter, mu
          fLight = log(fLight) - (pow(m_fSigma, 2)/2);

          //Do a random draw on the lognormal PDF
          fLight = clModelMath::LognormalRandomDraw(fLight, m_fSigma);

          if (fLight < 0) fLight = 0;
          if (fLight > 100) fLight = 100;

          //Assign GLI to the grid
          mp_oLightGrid->SetValueOfCell( iX, iY, m_iGridLightCode, fLight );

          //Assign the new basal areas to the grid
          mp_oLightGrid->SetValueOfCell( iX, iY, m_iGridConBACode, p_fConBAGrid[iX][iY] );
          mp_oLightGrid->SetValueOfCell( iX, iY, m_iGridAngBACode, p_fAngBAGrid[iX][iY] );
        }
      }
    }

    //Delete our temporary density grids
    for (iX = 0; iX < iNumXCells; iX++) {
      delete[] p_fConBAGrid[iX];
      delete[] p_fAngBAGrid[iX];
    }

    delete[] p_fConBAGrid;
    delete[] p_fAngBAGrid;

    m_bGridUpdated = true;
  }
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clBasalAreaLight::Action" );
    throw( stcErr );
  }
}
