//---------------------------------------------------------------------------
#ifndef HeightGLIWeibullMortalityH
#define HeightGLIWeibullMortalityH
//---------------------------------------------------------------------------
#include "MortalityBase.h"

/**
* Height-GLI Weibull Mortality - Version 1.0
*
* This evaluates mortality according to a weibull function of tree height
* and GLI.
*
* The equation used in this behavior is:
* <center>p = M<sub>max</sub> * exp(-a * H <sup>b</sup> - c * GLI <sup>d</sup>)</center>
* where
* <ul>
* <li>p = annual probability of mortality</li>
* <li>M<sub>max</sub> = max mortality parameter
* <li>a, b, c, d = parameters</li>
* <li>H = tree height in meters</li>
* <li>GLI, between 0 and 100</li>
* </ul>
*
* This class's namestring is "height gli weibull mortshell".  The parameter
* file call string is "Height GLI Weibull Mortality".
*
* Copyright 2006 Charles D. Canham.
* @author Lora E. Murphy
*
* <br>Edit history:
* <br>-----------------
* <br>January 26, 2006 - Created (LEM)
*/
class clHeightGLIWeibullMortality : virtual public clMortalityBase {
//note: need the virtual keyword to avoid base class ambiguity.

  public:

  /**
  * Constructor.
  *
  * @param p_oSimManager Sim Manager object.
  */
  clHeightGLIWeibullMortality(clSimManager *p_oSimManager);

  /**
  * Destructor.
  */
  ~clHeightGLIWeibullMortality();

  /**
  * Reads in values from the parameter file, then calls
  * GetTreeDataMemberCodes().
  *
  * @param p_oDoc DOM tree of parsed input file.
  * @throws modelErr if the max mortality parameter is not between 0 and 1.
  */
  void DoShellSetup(xercesc::DOMDocument *p_oDoc);

  /**
  * Calculates mortality according to the equation above.  Then uses a random
  * number to decide if a tree dies.
  *
  * @param p_oTree Tree being evaluated
  * @param fDbh Tree's DBH
  * @param iSpecies Species of the tree being evaluated
  * @return True if the tree is to die, false if it lives.
  */
  bool DoMort(clTree *p_oTree, const float &fDbh, const short int &iSpecies);

  protected:

  /** Return codes for the "Light" tree float data member variable.  Array
  * size is number of species to which this behavior applies by number of types
  */
  short int **mp_iLightCodes;

  /** Return codes for the "Height" tree float data member variable.  Array
  * size is number of species to which this behavior applies by number of
  * types*/
  short int **mp_iHeightCodes;

  /** Max mortality - sized number of behavior species*/
  float *mp_fMaxMort;

  /** Mortality "a" parameter - sized number of behavior species*/
  float *mp_fA;

  /** Mortality "b" parameter - sized number of behavior species*/
  float *mp_fB;

  /** Mortality "c" parameter - sized number of behavior species*/
  float *mp_fC;

  /** Mortality "d" parameter - sized number of behavior species*/
  float *mp_fD;

  /** To help access the other arrays*/
  int *mp_iIndexes;

  /** Number of years per timestep - for converting probabilities for multi-
  * year timesteps */
  float m_fYearsPerTimestep;

  /**
  * Queries for the return codes of the "Light" and "Height" tree data members.
  * Return codes are captured in the mp_iLightCodes and mp_iHeightCodes arrays.
  * @throws modelErr if any data member codes are missing.
  */
  void GetTreeDataMemberCodes();
};
//---------------------------------------------------------------------------
#endif
