//---------------------------------------------------------------------------
#include "Output.h"
#include "GridBase.h"
#include "SimManager.h"
#include "Plot.h"
#include "TreePopulation.h"
#include "ParsingFunctions.h"
#include "PlatformFuncs.h"

const int BUFFER_SIZE = 1023;

/////////////////////////////////////////////////////////////////////////////
// Destructor
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
// October 24, 2005 - Changed the tree save structure (LEM)
//////////////////////////////////////////////////////////////////////////////
clOutput::~clOutput()
{
  int i, j; //loop counter

  //Delete arrays
  if ( mp_treeSettings )
  {
    for ( i = 0; i < m_iNumSpecies; i++ )
    {
      for ( j = 0; j < m_iNumTypes; j++ )
      {
        delete[] mp_treeSettings[i] [j].p_iIntCodes;
        delete[] mp_treeSettings[i] [j].p_iFloatCodes;
        delete[] mp_treeSettings[i] [j].p_iCharCodes;
        delete[] mp_treeSettings[i] [j].p_iBoolCodes;
      }
      delete[] mp_treeSettings[i];
    }
    delete[] mp_treeSettings; mp_treeSettings = NULL;
  }

  if ( mp_gridSettings )
  {
    for ( i = 0; i < m_iNumGridsToSave; i++ )
    {
      delete[] mp_gridSettings[i].p_iIntCodes;
      delete[] mp_gridSettings[i].p_iFloatCodes;
      delete[] mp_gridSettings[i].p_iCharCodes;
      delete[] mp_gridSettings[i].p_iBoolCodes;
      delete[] mp_gridSettings[i].p_iPackageIntCodes;
      delete[] mp_gridSettings[i].p_iPackageFloatCodes;
      delete[] mp_gridSettings[i].p_iPackageCharCodes;
      delete[] mp_gridSettings[i].p_iPackageBoolCodes;
    }
    delete[] mp_gridSettings;
  } 
}

/////////////////////////////////////////////////////////////////////////////
// Constructor
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
/////////////////////////////////////////////////////////////////////////////
clOutput::clOutput( clSimManager * p_oSimManager ) : clWorkerBase( p_oSimManager ), clBehaviorBase( p_oSimManager )
{
  try
  {
    strcpy( m_cNameString, "output" );

    //Versions
    m_fVersionNumber = 1;
    m_fMinimumVersionNumber = 1;

    //Null pointers, initialize variables to empty
    mp_treeSettings = NULL;
    mp_gridSettings = NULL;
    m_iNumGridsToSave = 0;
    strcpy( m_cFileRoot, "" );

    //Allowed file types
    m_iNumAllowedTypes = 2;
    mp_iAllowedFileTypes = new int[m_iNumAllowedTypes];
    mp_iAllowedFileTypes[0] = parfile;
    mp_iAllowedFileTypes[1] = detailed_output;
  }
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clOutput::clOutput" );
    throw( stcErr );
  }
}


/////////////////////////////////////////////////////////////////////////////
// GetData()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
/////////////////////////////////////////////////////////////////////////////
void clOutput::GetData( DOMDocument * p_oDoc )
{
  try
  {
    clTreePopulation * p_oPop = ( clTreePopulation * ) mp_oSimManager->GetPopulationObject( "treepopulation" );;
    DOMElement * p_oElement; //for casting to DOM_Element
    char cTemp[MAX_FILENAME_SIZE];
    int i, j, k, iNumTimesteps, iMaxRootFileSize, iStrLen, iMaxSubplotName;
    bool bTest;
    
    //Get the number of cells in each direction
    m_iNumXCells = p_oPop->GetNumXCells();
    m_iNumYCells = p_oPop->GetNumYCells();
    
    ExtractTreeInfo( p_oDoc );
    ExtractGridInfo( p_oDoc );
    ExtractSubplotInfo( p_oDoc );

    //General data and flags

    //Filenames
    //get the doc root element
    p_oElement = p_oDoc->getDocumentElement();
    FillSingleValue( p_oElement, "ou_filename", cTemp, MAX_FILENAME_SIZE, true );

    strcpy( m_cFileRoot, cTemp );

    //Strip the filename of the tarball extension, if it's there, to get the
    //file root
    //Start by making cTemp all lower case
    iStrLen = strlen( cTemp );
    for ( i = 0; i < iStrLen; i++ )
    {
      cTemp[i] = tolower( cTemp[i] );
    }

    //Work backwards the length of the file extension in cTemp and compare
    if ( strlen( cTemp ) > strlen( TARBALL_FILE_EXT ) )
    {

      bTest = true;
      j = 0;
      iStrLen = strlen( cTemp );
      for ( i = iStrLen - strlen( TARBALL_FILE_EXT ); i < iStrLen; i++ )
      {
        if ( cTemp[i] != TARBALL_FILE_EXT[j] )
        {
          bTest = false;
          break;
        }
        j++;
      }

      if ( bTest == true )
      {

        //The file extension's already on there - so strip the extension
        //off of root
        strcpy( m_cTarball, m_cFileRoot );
        m_cFileRoot[strlen( m_cFileRoot ) - strlen( TARBALL_FILE_EXT )] = '\0';

      }
    }

    //We want all our filenames to fit into MAX_FILENAME_SIZE.  So we have to
    //make sure our root is short enough to accommodate the timestep and subplot 
    //tack-ons.  Get the number of timesteps and calculate the max length of 
    //our root filename
    iNumTimesteps = mp_oSimManager->GetNumberOfTimesteps();
    if (0 == m_iNumSubplotsToSave) {
      //No subplots, so the max is filename length minus number of digits in
      //max timesteps minus 1 for the underscore minus the length of the 
      //zipped XML extension (The -2 is one for the underscore and one because
      //log10 gives us 1 less than the number of digits in timesteps)
      iMaxRootFileSize = MAX_FILENAME_SIZE - (int)floor( log10( iNumTimesteps ) ) - 2 - strlen( ZIPPED_FILE_EXT ) - 1; //one for the NULL
    } else {
      //Subplots, so the max is filename length minus number of digits in
      //max timesteps minus 2 for the underscore minus the longest subplot name
      //minus the length of the zipped XML extension (The -3 is for the 
      //underscores and one because log10 gives us 1 less than the number of 
      //digits in timesteps)
      
      //Get the max subplot name length
      iMaxSubplotName = 0;
      for (i = 0; i < m_iNumSubplotsToSave; i++) {
        iStrLen = strlen( mp_subplots[i].cSubplotName );
        if (iStrLen > iMaxSubplotName) iMaxSubplotName = iStrLen;
      }
      iMaxRootFileSize = MAX_FILENAME_SIZE - (int)floor( log10( iNumTimesteps ) ) - 3 - iMaxSubplotName - strlen( ZIPPED_FILE_EXT ) - 1; //one for the NULL      
    }

    //If our file length is over that value, throw an error.
    iStrLen = strlen( m_cFileRoot );
    if ( iStrLen > iMaxRootFileSize )
    {
      modelErr stcErr;
      stcErr.iErrorCode = BAD_FILE;
      strcpy( stcErr.cFunction, "clOutput::GetData" );
      strcpy( stcErr.cMoreInfo, "Detailed output file name is too long." );
      throw( stcErr );
    }

    //Make the tarball file name by appending the tarball extension to the root
    sprintf( m_cTarball, "%s%s", m_cFileRoot, TARBALL_FILE_EXT );
    
    //Make the tarball names of all the subplots
    for (i = 0; i < m_iNumSubplotsToSave; i++) {
      sprintf( mp_subplots[i].cSubplotTarball, "%s%s%s%s", m_cFileRoot, "_", 
                  mp_subplots[i].cSubplotName, TARBALL_FILE_EXT );
    }

    //Now take off the directory structure so that TAR won't preserve a
    //directory structure for files - which is good since very long filenames
    //confound TAR otherwise
    strcpy( cTemp, m_cFileRoot );
    iStrLen = strlen( m_cFileRoot ) - 1;
    for ( i = iStrLen; i >= 0; i-- )
    {
      if ( '\\' == cTemp[i] || '/' == cTemp[i] )
      {
        k = 0;
        iStrLen++;
        for ( j = i + 1; j < iStrLen; j++ )
        {
          m_cFileRoot[k] = cTemp[j];
          k++;
        }
        m_cFileRoot[k] = '\0';
        break;
      }
    }

    //Just double-check that we have settings to save - if not, erase the
    //filename
    bTest = false;
    for ( i = 0; i < m_iNumSpecies; i++ )
      for ( j = 0; j < m_iNumTypes; j++ )
        if ( mp_treeSettings[i] [j].iSaveFreq > -1 )
        {
          bTest = true; break;
        }
    if ( false == bTest && 0 == m_iNumGridsToSave )
      strcpy( m_cFileRoot, "" );

    //Make sure that everything's in place for zipping - if not throw error
    if ( !TarballSetup( mp_oSimManager->GetAppPath() ) )
    {
      modelErr stcErr;
      stcErr.iErrorCode = CANT_FIND_OBJECT;
      sprintf( stcErr.cMoreInfo, "%s%s", "Missing file gzip.exe and/or tar.exe in path ", mp_oSimManager->GetAppPath() );
      strcpy( stcErr.cFunction, "clOutput::GetData" );
      throw( stcErr );
    }

    //Write the output file headers
    WriteDetailedOutputHeader();

    //Write the initial conditions timestep
    Action();

  } //end of try block
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clOutput::GetData" );
    throw( stcErr );
  }
}


/////////////////////////////////////////////////////////////////////////////
// ExtractTreeInfo()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
// October 24, 2005 - Changed the tree save structure (LEM)
/////////////////////////////////////////////////////////////////////////////
void clOutput::ExtractTreeInfo( DOMDocument * p_oDoc )
{
  try
  {
    clTreePopulation * p_oPop = ( clTreePopulation * ) mp_oSimManager->GetPopulationObject( "treepopulation" );;
    DOMNodeList * p_oSettingsList, //holds the list of tree info structures
         * p_oDataMemberList; //holds list of tree data members
    DOMNode * p_oDocNode;
    DOMElement * p_oTreeInfo, //one tree type's info
         * p_oDataMember; //one data member
    char * cData; //for extracting data from the file
    short int iNumSettings, //how many types we have output data for in the file
         iSpecies, iType, //species number
         iDataMemberCode, //data member code for a label
         i, j; //loop counters

    //Get the tree population
    m_iNumSpecies = p_oPop->GetNumberOfSpecies();
    m_iNumTypes = p_oPop->GetNumberOfTypes();

    // Look for tree info
    //Search for all tree type info sets
    p_oSettingsList = p_oDoc->getElementsByTagName( XMLString::transcode( "ou_treeInfo" ) );
    iNumSettings = p_oSettingsList->getLength();

    //Allocate the tree memory and initialize - set frequencies to -1 so we
    //can distinguish absence of data from the user setting 0
    mp_treeSettings = new stcTreeOutputInfo * [m_iNumSpecies];
    for ( i = 0; i < m_iNumSpecies; i++ )
    {
      mp_treeSettings[i] = new stcTreeOutputInfo[m_iNumTypes];
      for ( j = 0; j < m_iNumTypes; j++ )
      {
        mp_treeSettings[i] [j].iSaveFreq = -1;
        mp_treeSettings[i] [j].iSumTimestep = -1;
        mp_treeSettings[i] [j].iNumInts = 0;
        mp_treeSettings[i] [j].iNumFloats = 0;
        mp_treeSettings[i] [j].iNumChars = 0;
        mp_treeSettings[i] [j].iNumBools = 0;
        mp_treeSettings[i] [j].bSaveThisTimeStep = false;
        mp_treeSettings[i] [j].p_iIntCodes = NULL;
        mp_treeSettings[i] [j].p_iFloatCodes = NULL;
        mp_treeSettings[i] [j].p_iCharCodes = NULL;
        mp_treeSettings[i] [j].p_iBoolCodes = NULL;
      }
    }

    //If we found nothing, exit with no error
    if ( iNumSettings == 0 ) return;

    //go through each type and break out its info
    for ( i = 0; i < iNumSettings; i++ )
    {
      p_oDocNode = p_oSettingsList->item( i );
      p_oTreeInfo = ( DOMElement * ) p_oDocNode;

      //get the type name - if something goes wrong throw an error
      cData = XMLString::transcode( p_oTreeInfo->getAttributeNode( XMLString::transcode( "type" ) )->getNodeValue() );
      if ( strcmp( "Seed", cData ) == 0 )
        iType = clTreePopulation::seed;
      else if ( strcmp( "Seedling", cData ) == 0 )
        iType = clTreePopulation::seedling;
      else if ( strcmp( "Sapling", cData ) == 0 )
        iType = clTreePopulation::sapling;
      else if ( strcmp( "Adult", cData ) == 0 )
        iType = clTreePopulation::adult;
      else if ( strcmp( "Snag", cData ) == 0 )
        iType = clTreePopulation::snag;
      else if ( strcmp( "Woody_debris", cData ) == 0 )
        iType = clTreePopulation::woody_debris;
      else
      {
        modelErr stcErr;
        strcpy( stcErr.cFunction, "clOutput::ExtractTreeInfo" );
        stcErr.iErrorCode = BAD_DATA;
        sprintf( stcErr.cMoreInfo, "%s%s", "unrecognized type - ", cData );
        throw( stcErr );
      }

      //Species
      cData = XMLString::transcode( p_oTreeInfo->getAttributeNode( XMLString::transcode( "species" ) )->getNodeValue() );
      iSpecies = p_oPop->TranslateSpeciesNameToCode( cData );
      if ( -1 == iSpecies )
      { //unrecognized species - throw error
        modelErr stcErr;
        stcErr.iErrorCode = BAD_DATA;
        strcpy( stcErr.cFunction, "clOutput::ExtractTreeInfo" );
        sprintf( stcErr.cMoreInfo, "%s%s", "Unrecognized species - ", cData );
        throw( stcErr );
      }

      //Frequency data
      FillSingleValue( p_oTreeInfo, "ou_saveFreq", ( int * ) & mp_treeSettings[iSpecies] [iType].iSaveFreq, true );

      //Make sure that the frequency is not negative - if so, throw error
      if ( mp_treeSettings[iSpecies] [iType].iSaveFreq < 0 )
      {
        modelErr stcErr;
        stcErr.iErrorCode = BAD_DATA;
        strcpy( stcErr.cFunction, "clOutput::ExtractTreeInfo" );
        strcpy( stcErr.cMoreInfo, "Save frequency is negative" );
        throw( stcErr );
      }

      //Number to save for each data member type
      //****************Ints*******************
      p_oDataMemberList = p_oTreeInfo->getElementsByTagName( XMLString::transcode( "ou_int" ) );
      if ( 0 != p_oDataMemberList->getLength() )
      {
        mp_treeSettings[iSpecies] [iType].iNumInts = p_oDataMemberList->getLength();
        //Declare the int array
        mp_treeSettings[iSpecies] [iType].p_iIntCodes = new short int[mp_treeSettings[iSpecies] [iType].iNumInts];
        for ( j = 0; j < mp_treeSettings[iSpecies] [iType].iNumInts; j++ )
        {
          p_oDocNode = p_oDataMemberList->item( j );
          p_oDataMember = ( DOMElement * ) p_oDocNode;
          cData = XMLString::transcode( p_oDataMember->getFirstChild()->getNodeValue() );
          iDataMemberCode = p_oPop->GetIntDataCode( cData, iSpecies, iType );
          if ( -1 == iDataMemberCode )
          { //unrecognized data code - throw error
            modelErr stcErr;
            stcErr.iErrorCode = BAD_DATA;
            strcpy( stcErr.cFunction, "clOutput::ExtractTreeInfo" );
            sprintf( stcErr.cMoreInfo, "%s%s", "Unrecognized int data member label - ", cData );
            throw( stcErr );
          }
          mp_treeSettings[iSpecies] [iType].p_iIntCodes[j] = iDataMemberCode;
        }
      } //end of if (0 != p_oDataMemberList->getLength())
      //****************Floats*******************
      p_oDataMemberList = p_oTreeInfo->getElementsByTagName( XMLString::transcode( "ou_float" ) );
      if ( 0 != p_oDataMemberList->getLength() )
      {
        mp_treeSettings[iSpecies] [iType].iNumFloats = p_oDataMemberList->getLength();
        //Declare the float array
        mp_treeSettings[iSpecies] [iType].p_iFloatCodes = new short int[mp_treeSettings[iSpecies] [iType].iNumFloats];
        for ( j = 0; j < mp_treeSettings[iSpecies] [iType].iNumFloats; j++ )
        {
          p_oDocNode = p_oDataMemberList->item( j );
          p_oDataMember = ( DOMElement * ) p_oDocNode;
          cData = XMLString::transcode( p_oDataMember->getFirstChild()->getNodeValue() );
          iDataMemberCode = p_oPop->GetFloatDataCode( cData, iSpecies, iType );
          if ( -1 == iDataMemberCode )
          { //unrecognized data code - throw error
            modelErr stcErr;
            stcErr.iErrorCode = BAD_DATA;
            strcpy( stcErr.cFunction, "clOutput::ExtractTreeInfo" );
            sprintf( stcErr.cMoreInfo, "%s%s", "Unrecognized float data member label - ", cData );
            throw( stcErr );
          }
          mp_treeSettings[iSpecies] [iType].p_iFloatCodes[j] = iDataMemberCode;
        }
      } //end of if (0 != p_oDataMemberList->getLength())
      //****************Chars*******************
      p_oDataMemberList = p_oTreeInfo->getElementsByTagName( XMLString::transcode( "ou_char" ) );
      if ( 0 != p_oDataMemberList->getLength() )
      {
        mp_treeSettings[iSpecies] [iType].iNumChars = p_oDataMemberList->getLength();
        //Declare the int array
        mp_treeSettings[iSpecies] [iType].p_iCharCodes = new short int[mp_treeSettings[iSpecies] [iType].iNumChars];
        for ( j = 0; j < mp_treeSettings[iSpecies] [iType].iNumChars; j++ )
        {
          p_oDocNode = p_oDataMemberList->item( j );
          p_oDataMember = ( DOMElement * ) p_oDocNode;
          cData = XMLString::transcode( p_oDataMember->getFirstChild()->getNodeValue() );
          iDataMemberCode = p_oPop->GetCharDataCode( cData, iSpecies, iType );
          if ( -1 == iDataMemberCode )
          { //unrecognized data code - throw error
            modelErr stcErr;
            stcErr.iErrorCode = BAD_DATA;
            strcpy( stcErr.cFunction, "clOutput::ExtractTreeInfo" );
            sprintf( stcErr.cMoreInfo, "%s%s", "Unrecognized char data member label - ", cData );
            throw( stcErr );
          }
          mp_treeSettings[iSpecies] [iType].p_iCharCodes[j] = iDataMemberCode;
        }
      } //end of if (0 != p_oDataMemberList->getLength())
      //****************Bools*******************
      p_oDataMemberList = p_oTreeInfo->getElementsByTagName( XMLString::transcode( "ou_bool" ) );
      if ( 0 != p_oDataMemberList->getLength() )
      {
        mp_treeSettings[iSpecies] [iType].iNumBools = p_oDataMemberList->getLength();
        //Declare the int array
        mp_treeSettings[iSpecies] [iType].p_iBoolCodes = new short int[mp_treeSettings[iSpecies] [iType].iNumBools];
        for ( j = 0; j < mp_treeSettings[iSpecies] [iType].iNumBools; j++ )
        {
          p_oDocNode = p_oDataMemberList->item( j );
          p_oDataMember = ( DOMElement * ) p_oDocNode;
          cData = XMLString::transcode( p_oDataMember->getFirstChild()->getNodeValue() );
          iDataMemberCode = p_oPop->GetBoolDataCode( cData, iSpecies, iType );
          if ( -1 == iDataMemberCode )
          { //unrecognized data code - throw error
            modelErr stcErr;
            stcErr.iErrorCode = BAD_DATA;
            strcpy( stcErr.cFunction, "clOutput::ExtractTreeInfo" );
            sprintf( stcErr.cMoreInfo, "%s%s", "Unrecognized bool data member label - ", cData );
            throw( stcErr );
          }
          mp_treeSettings[iSpecies] [iType].p_iBoolCodes[j] = iDataMemberCode;
        }
      } //end of if (0 != p_oDataMemberList->getLength())
    } //end of for (i = 0; i < m_iNumTypeToSave; i++)
  } //end of try block
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clOutput::ExtractTreeInfo" );
    throw( stcErr );
  }
}


/////////////////////////////////////////////////////////////////////////////
// ExtractGridInfo()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
/////////////////////////////////////////////////////////////////////////////
void clOutput::ExtractGridInfo( DOMDocument * p_oDoc )
{
  try
  {
    DOMNodeList * p_oGridList, //list of grids
         * p_oDataMemberList; //list of data members
    DOMNode * p_oDocNode; //for retrieving data
    DOMElement * p_oGrid, //one grid element
         * p_oDataMember; //data member element
    char * cData; //for extracting text data
    short int iNumGrids, //number of grids found
         iDataMemberCode, //data member code
         i, j; //loop counters

    //Get all the grid info structures out of the file, if any
    p_oGridList = p_oDoc->getElementsByTagName( XMLString::transcode( "ou_gridInfo" ) );
    iNumGrids = p_oGridList->getLength();
    if ( 0 == iNumGrids ) return; //if nothing found exit without error

    //Allocate memory - set frequencies to -1 so we
    //can distinguish absence of data from the user setting 0
    m_iNumGridsToSave = iNumGrids;
    mp_gridSettings = new stcGridOutputInfo[m_iNumGridsToSave];
    for ( i = 0; i < m_iNumGridsToSave; i++ )
    {
      mp_gridSettings[i].iSaveFreq = -1;
      mp_gridSettings[i].iSumTimestep = 0;
      mp_gridSettings[i].iNumInts = 0;
      mp_gridSettings[i].iNumFloats = 0;
      mp_gridSettings[i].iNumChars = 0;
      mp_gridSettings[i].iNumBools = 0;
      mp_gridSettings[i].p_iIntCodes = NULL;
      mp_gridSettings[i].p_iFloatCodes = NULL;
      mp_gridSettings[i].p_iCharCodes = NULL;
      mp_gridSettings[i].p_iBoolCodes = NULL;
      mp_gridSettings[i].iNumPackageInts = 0;
      mp_gridSettings[i].iNumPackageFloats = 0;
      mp_gridSettings[i].iNumPackageChars = 0;
      mp_gridSettings[i].iNumPackageBools = 0;
      mp_gridSettings[i].p_iPackageIntCodes = NULL;
      mp_gridSettings[i].p_iPackageFloatCodes = NULL;
      mp_gridSettings[i].p_iPackageCharCodes = NULL;
      mp_gridSettings[i].p_iPackageBoolCodes = NULL;
    } //end of for (i = 0; i < m_iNumGridsToSave; i++)

    //Go through the grid objects in the file one at a time and extract the info
    for ( i = 0; i < m_iNumGridsToSave; i++ )
    {
      p_oDocNode = p_oGridList->item( i );
      p_oGrid = ( DOMElement * ) p_oDocNode;

      //First make sure that this is for a valid grid object
      cData = XMLString::transcode( p_oGrid->getAttributeNode( XMLString::transcode( "gridName" ) )->getNodeValue() );
      mp_gridSettings[i].p_oGridPointer = mp_oSimManager->GetGridObject( cData );
      if ( NULL == mp_gridSettings[i].p_oGridPointer )
      {
        modelErr stcErr;
        strcpy( stcErr.cFunction, "clOutput::ExtractGridInfo" );
        stcErr.iErrorCode = CANT_FIND_OBJECT;
        strcpy( stcErr.cMoreInfo, cData );
        throw( stcErr );
      }

      //Get frequency
      FillSingleValue( p_oGrid, "ou_saveFreq", ( int * ) & mp_gridSettings[i].iSaveFreq, true );

      //Number to save for each data member type
      //****************Ints*******************
      p_oDataMemberList = p_oGrid->getElementsByTagName( XMLString::transcode( "ou_int" ) );
      if ( 0 != p_oDataMemberList->getLength() )
      {
        mp_gridSettings[i].iNumInts = p_oDataMemberList->getLength();
        //Declare the int array
        mp_gridSettings[i].p_iIntCodes = new short int[mp_gridSettings[i].iNumInts];
        for ( j = 0; j < mp_gridSettings[i].iNumInts; j++ )
        {
          p_oDocNode = p_oDataMemberList->item( j );
          p_oDataMember = ( DOMElement * ) p_oDocNode;
          cData = XMLString::transcode( p_oDataMember->getFirstChild()->getNodeValue() );
          iDataMemberCode = mp_gridSettings[i].p_oGridPointer->GetIntDataCode( cData );
          if ( -1 == iDataMemberCode )
          { //unrecognized data code - throw error
            modelErr stcErr;
            stcErr.iErrorCode = BAD_DATA;
            strcpy( stcErr.cFunction, "clOutput::ExtractGridInfo" );
            sprintf( stcErr.cMoreInfo, "%s%s", "Unrecognized int data member label - ", cData );
            throw( stcErr );
          }
          mp_gridSettings[i].p_iIntCodes[j] = iDataMemberCode;
        }
      } //end of if (0 != p_oDataMemberList->getLength())
      //****************Floats*******************
      p_oDataMemberList = p_oGrid->getElementsByTagName( XMLString::transcode( "ou_float" ) );
      if ( 0 != p_oDataMemberList->getLength() )
      {
        mp_gridSettings[i].iNumFloats = p_oDataMemberList->getLength();
        //Declare the float array
        mp_gridSettings[i].p_iFloatCodes = new short int[mp_gridSettings[i].iNumFloats];
        for ( j = 0; j < mp_gridSettings[i].iNumFloats; j++ )
        {
          p_oDocNode = p_oDataMemberList->item( j );
          p_oDataMember = ( DOMElement * ) p_oDocNode;
          cData = XMLString::transcode( p_oDataMember->getFirstChild()->getNodeValue() );
          iDataMemberCode = mp_gridSettings[i].p_oGridPointer->GetFloatDataCode( cData );
          if ( -1 == iDataMemberCode )
          { //unrecognized data code - throw error
            modelErr stcErr;
            stcErr.iErrorCode = BAD_DATA;
            strcpy( stcErr.cFunction, "clOutput::ExtractGridInfo" );
            sprintf( stcErr.cMoreInfo, "%s%s", "Unrecognized float data member label - ", cData );
            throw( stcErr );
          }
          mp_gridSettings[i].p_iFloatCodes[j] = iDataMemberCode;
        }
      } //end of if (0 != p_oDataMemberList->getLength())
      //****************Chars*******************
      p_oDataMemberList = p_oGrid->getElementsByTagName( XMLString::transcode( "ou_char" ) );
      if ( 0 != p_oDataMemberList->getLength() )
      {
        mp_gridSettings[i].iNumChars = p_oDataMemberList->getLength();
        //Declare the char array
        mp_gridSettings[i].p_iCharCodes = new short int[mp_gridSettings[i].iNumChars];
        for ( j = 0; j < mp_gridSettings[i].iNumChars; j++ )
        {
          p_oDocNode = p_oDataMemberList->item( j );
          p_oDataMember = ( DOMElement * ) p_oDocNode;
          cData = XMLString::transcode( p_oDataMember->getFirstChild()->getNodeValue() );
          iDataMemberCode = mp_gridSettings[i].p_oGridPointer->GetCharDataCode( cData );
          if ( -1 == iDataMemberCode )
          { //unrecognized data code - throw error
            modelErr stcErr;
            stcErr.iErrorCode = BAD_DATA;
            strcpy( stcErr.cFunction, "clOutput::ExtractGridInfo" );
            sprintf( stcErr.cMoreInfo, "%s%s", "Unrecognized char data member label - ", cData );
            throw( stcErr );
          }
          mp_gridSettings[i].p_iCharCodes[j] = iDataMemberCode;
        }
      } //end of if (0 != p_oDataMemberList->getLength())
      //****************Bools*******************
      p_oDataMemberList = p_oGrid->getElementsByTagName( XMLString::transcode( "ou_bool" ) );
      if ( 0 != p_oDataMemberList->getLength() )
      {
        mp_gridSettings[i].iNumBools = p_oDataMemberList->getLength();
        //Declare the bool array
        mp_gridSettings[i].p_iBoolCodes = new short int[mp_gridSettings[i].iNumBools];
        for ( j = 0; j < mp_gridSettings[i].iNumBools; j++ )
        {
          p_oDocNode = p_oDataMemberList->item( j );
          p_oDataMember = ( DOMElement * ) p_oDocNode;
          cData = XMLString::transcode( p_oDataMember->getFirstChild()->getNodeValue() );
          iDataMemberCode = mp_gridSettings[i].p_oGridPointer->GetBoolDataCode( cData );
          if ( -1 == iDataMemberCode )
          { //unrecognized data code - throw error
            modelErr stcErr;
            stcErr.iErrorCode = BAD_DATA;
            strcpy( stcErr.cFunction, "clOutput::ExtractGridInfo" );
            sprintf( stcErr.cMoreInfo, "%s%s", "Unrecognized bool data member label - ", cData );
            throw( stcErr );
          }
          mp_gridSettings[i].p_iBoolCodes[j] = iDataMemberCode;
        }
      } //end of if (0 != p_oDataMemberList->getLength())


      //Packages
      //****************Package Ints*******************
      p_oDataMemberList = p_oGrid->getElementsByTagName( XMLString::transcode( "ou_packageInt" ) );
      if ( 0 != p_oDataMemberList->getLength() )
      {
        mp_gridSettings[i].iNumPackageInts = p_oDataMemberList->getLength();
        //Declare the int array
        mp_gridSettings[i].p_iPackageIntCodes = new short int[mp_gridSettings[i].iNumPackageInts];
        for ( j = 0; j < mp_gridSettings[i].iNumPackageInts; j++ )
        {
          p_oDocNode = p_oDataMemberList->item( j );
          p_oDataMember = ( DOMElement * ) p_oDocNode;
          cData = XMLString::transcode( p_oDataMember->getFirstChild()->getNodeValue() );
          iDataMemberCode = mp_gridSettings[i].p_oGridPointer->GetPackageIntDataCode( cData );
          if ( -1 == iDataMemberCode )
          { //unrecognized data code - throw error
            modelErr stcErr;
            stcErr.iErrorCode = BAD_DATA;
            strcpy( stcErr.cFunction, "clOutput::ExtractGridInfo" );
            sprintf( stcErr.cMoreInfo, "%s%s", "Unrecognized package int data member label - ", cData );
            throw( stcErr );
          }
          mp_gridSettings[i].p_iPackageIntCodes[j] = iDataMemberCode;
        }
      } //end of if (0 != p_oDataMemberList->getLength())
      //****************Package Floats*******************
      p_oDataMemberList = p_oGrid->getElementsByTagName( XMLString::transcode( "ou_packageFloat" ) );
      if ( 0 != p_oDataMemberList->getLength() )
      {
        mp_gridSettings[i].iNumPackageFloats = p_oDataMemberList->getLength();
        //Declare the float array
        mp_gridSettings[i].p_iPackageFloatCodes = new short int[mp_gridSettings[i].iNumPackageFloats];
        for ( j = 0; j < mp_gridSettings[i].iNumPackageFloats; j++ )
        {
          p_oDocNode = p_oDataMemberList->item( j );
          p_oDataMember = ( DOMElement * ) p_oDocNode;
          cData = XMLString::transcode( p_oDataMember->getFirstChild()->getNodeValue() );
          iDataMemberCode = mp_gridSettings[i].p_oGridPointer->GetPackageFloatDataCode( cData );
          if ( -1 == iDataMemberCode )
          { //unrecognized data code - throw error
            modelErr stcErr;
            stcErr.iErrorCode = BAD_DATA;
            strcpy( stcErr.cFunction, "clOutput::ExtractGridInfo" );
            sprintf( stcErr.cMoreInfo, "%s%s", "Unrecognized package float data member label - ", cData );
            throw( stcErr );
          }
          mp_gridSettings[i].p_iPackageFloatCodes[j] = iDataMemberCode;
        }
      } //end of if (0 != p_oDataMemberList->getLength())
      //****************Package Chars*******************
      p_oDataMemberList = p_oGrid->getElementsByTagName( XMLString::transcode( "ou_packageChar" ) );
      if ( 0 != p_oDataMemberList->getLength() )
      {
        mp_gridSettings[i].iNumPackageChars = p_oDataMemberList->getLength();
        //Declare the char array
        mp_gridSettings[i].p_iPackageCharCodes = new short int[mp_gridSettings[i].iNumPackageChars];
        for ( j = 0; j < mp_gridSettings[i].iNumPackageChars; j++ )
        {
          p_oDocNode = p_oDataMemberList->item( j );
          p_oDataMember = ( DOMElement * ) p_oDocNode;
          cData = XMLString::transcode( p_oDataMember->getFirstChild()->getNodeValue() );
          iDataMemberCode = mp_gridSettings[i].p_oGridPointer->GetPackageCharDataCode( cData );
          if ( -1 == iDataMemberCode )
          { //unrecognized data code - throw error
            modelErr stcErr;
            stcErr.iErrorCode = BAD_DATA;
            strcpy( stcErr.cFunction, "clOutput::ExtractGridInfo" );
            sprintf( stcErr.cMoreInfo, "%s%s", "Unrecognized package char data member label - ", cData );
            throw( stcErr );
          }
          mp_gridSettings[i].p_iPackageCharCodes[j] = iDataMemberCode;
        }
      } //end of if (0 != p_oDataMemberList->getLength())
      //****************Package Bools*******************
      p_oDataMemberList = p_oGrid->getElementsByTagName( XMLString::transcode( "ou_packageBool" ) );
      if ( 0 != p_oDataMemberList->getLength() )
      {
        mp_gridSettings[i].iNumPackageBools = p_oDataMemberList->getLength();
        //Declare the bool array
        mp_gridSettings[i].p_iPackageBoolCodes = new short int[mp_gridSettings[i].iNumPackageBools];
        for ( j = 0; j < mp_gridSettings[i].iNumPackageBools; j++ )
        {
          p_oDocNode = p_oDataMemberList->item( j );
          p_oDataMember = ( DOMElement * ) p_oDocNode;
          cData = XMLString::transcode( p_oDataMember->getFirstChild()->getNodeValue() );
          iDataMemberCode = mp_gridSettings[i].p_oGridPointer->GetPackageBoolDataCode( cData );
          if ( -1 == iDataMemberCode )
          { //unrecognized data code - throw error
            modelErr stcErr;
            stcErr.iErrorCode = BAD_DATA;
            strcpy( stcErr.cFunction, "clOutput::ExtractGridInfo" );
            sprintf( stcErr.cMoreInfo, "%s%s", "Unrecognized package bool data member label - ", cData );
            throw( stcErr );
          }
          mp_gridSettings[i].p_iPackageBoolCodes[j] = iDataMemberCode;
        }
      } //end of if (0 != p_oDataMemberList->getLength())
    } //end of (i = 0; i < m_iNumGridsToSave; i++)
  } //end of try block
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clOutput::ExtractGridInfo" );
    throw( stcErr );
  }
}

////////////////////////////////////////////////////////////////////////////
// ExtractSubplotInfo()
// Edit history:
// -------------
// May 10, 2006 - Created (LEM)
////////////////////////////////////////////////////////////////////////////
void clOutput::ExtractSubplotInfo( xercesc::DOMDocument * p_oDoc )
{
  try
  {
    DOMNodeList * p_oGridCellList, //for getting the list of grid cells to save for
         * p_oSubplotList; //list of subplots
    DOMNode * p_oDocNode; //for retrieving data
    DOMElement * p_oSubplot, //one subplot element
         * p_oElement; //generic element object
    int iX, iY; //grid cell X and Y coordinate numbers
    short int iNumSubplots, //number of subplots found
         iNumCells, //number of grid cells found for one subplot
         i, j; //loop counters
    bool bTest;
    
    m_iNumSubplotsToSave = 0;

    //If there was no tree data saved, we will ignore subplots anyway, so
    //exit
    bTest = false;
    for ( i = 0; i < m_iNumSpecies; i++ )
      for ( j = 0; j < m_iNumTypes; j++ )
        if ( mp_treeSettings[i] [j].iSaveFreq > -1 )
        {
          bTest = true; break;
        }
    if ( false == bTest ) return;

    //Get all the subplot info structures out of the file, if any
    p_oSubplotList = p_oDoc->getElementsByTagName( XMLString::transcode( "ou_subplot" ) );
    iNumSubplots = p_oSubplotList->getLength();
    if ( 0 == iNumSubplots ) return; //if nothing found exit without error

    //Allocate memory
    m_iNumSubplotsToSave = iNumSubplots;
    mp_subplots = new stcSubplotInfo[m_iNumSubplotsToSave];
    for ( i = 0; i < m_iNumSubplotsToSave; i++ )
    {
      mp_subplots[i].p_bUseCell = new bool*[m_iNumXCells];
      for (iX = 0; iX < m_iNumXCells; iX++) {
        mp_subplots[i].p_bUseCell[iX] = new bool[m_iNumYCells];
        for (iY = 0; iY < m_iNumYCells; iY++) {
          mp_subplots[i].p_bUseCell[iX][iY] = false; 
        } 
      }
    } //end of for (i = 0; i < m_iNumSubplotsToSave; i++)

    //Go through the subplots in the file one at a time and extract the info
    for ( i = 0; i < m_iNumSubplotsToSave; i++ )
    {
      p_oDocNode = p_oSubplotList->item( i );
      p_oSubplot = ( DOMElement * ) p_oDocNode;

      //Get subplot name
      FillSingleValue( p_oSubplot, "ou_subplotName", mp_subplots[i].cSubplotName, MAX_SUBPLOT_NAME_SIZE, true );

      //Get grid coords
      p_oGridCellList = p_oSubplot->getElementsByTagName( XMLString::transcode( "pointSet" ) );
      if ( 0 == p_oGridCellList->getLength() )
      {
        modelErr stcErr;
        stcErr.iErrorCode = DATA_MISSING;
        strcpy( stcErr.cFunction, "clOutput::ExtractSubplotInfo" );
        strcpy( stcErr.cMoreInfo, "pointSet" );
        throw( stcErr );
      }
      p_oDocNode = p_oGridCellList->item( 0 );
      p_oElement = ( DOMElement * ) p_oDocNode;
      p_oGridCellList = p_oElement->getElementsByTagName( XMLString::transcode( "po_point" ) );
      iNumCells = p_oGridCellList->getLength();
      if ( 0 == iNumCells )
      {
        modelErr stcErr;
        stcErr.iErrorCode = DATA_MISSING;
        strcpy( stcErr.cFunction, "clOutput::ExtractSubplotInfo" );
        strcpy( stcErr.cMoreInfo, "po_point" );
        throw( stcErr );
      }

      //Set each cell in the list to "true" in this subplot's cell array
      for ( j = 0; j < iNumCells; j++ )
      {
        p_oDocNode = p_oGridCellList->item( j );
        p_oElement = ( DOMElement * ) p_oDocNode;
        iX = atoi( XMLString::transcode( p_oElement->getAttributeNode( XMLString::transcode( "x" ) )->getNodeValue() ) );
        iY = atoi( XMLString::transcode( p_oElement->getAttributeNode( XMLString::transcode( "y" ) )->getNodeValue() ) );
        //Make sure that the values are within the plot
        if ( iX >= m_iNumXCells || iX < 0 || iY >= m_iNumYCells || iY < 0 )
        {
          modelErr stcErr;
          strcpy( stcErr.cFunction, "clOutput::ExtractSubplotInfo" );
          sprintf( stcErr.cMoreInfo, "%s%d%s%d", "Point value outside plot - ", iX, ", ", iY );
          stcErr.iErrorCode = BAD_DATA;
        }
        
        mp_subplots[i].p_bUseCell[iX][iY] = true;
    
      } //end of for (j = 0; j < iNumCells; j++)
    } //end of for (i = 0; i < m_iNumSubplotsToSave; i++)
  } //end of try block
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clOutput::ExtractSubplotInfo" );
    throw( stcErr );
  }
}


/////////////////////////////////////////////////////////////////////////////
// WriteDetailedOutputHeader()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
/////////////////////////////////////////////////////////////////////////////
void clOutput::WriteDetailedOutputHeader()
{
  const int BUFFER_SIZE = 128;

  FILE ** p_outFile = NULL, //array of detailed output file streams, one for each
                     //subplot plus one for the whole plot
       * inFile; //parameter file
  char cFileCode[9], //holds the file code
       ** p_cFilename = NULL, 
       cBuf[BUFFER_SIZE]; //for transferring between files
  char * cPos;
  int iNumRead, //for testing for end-of-file
      iStrLen, 
      iNumOutFiles = m_iNumSubplotsToSave + 1,
      iPos, i;
  bool bReading; //whether or not we found the <paramFile> tag
    
  try
  {

    if ( strlen( m_cFileRoot ) == 0 ) return;
    
    //Declare the array of out files and filenames
    p_outFile = new FILE*[iNumOutFiles];
    p_cFilename = new char*[iNumOutFiles];
    for (i = 0; i < iNumOutFiles; i++) {
      p_cFilename[i] = new char[MAX_FILENAME_SIZE];
    }

    //Tack on the file extension to the root for the whole-plot file
    sprintf( p_cFilename[0], "%s%s", m_cFileRoot, DETAILED_OUTPUT_FILE_EXT );

    //There is a possibility that the detailed output filename and the
    //parameter filename are the same - especially if a detailed output file
    //was loaded as a parameter file.  In this case, just skip out
    if ( strcmp( p_cFilename[0], mp_oSimManager->GetParFilename() ) == 0 )
      return;

    //Open detailed output file
    p_outFile[0] = fopen( p_cFilename[0], "wt" ); //overwrite existing

    //If we can't open the file throw an error
    if ( NULL == p_outFile[0] )
    {
      modelErr stcErr;
      stcErr.iErrorCode = BAD_FILE;
      strcpy( stcErr.cFunction, "clOutput::WriteDetailedOutputHeader" );
      sprintf( stcErr.cMoreInfo, "%s%s%s", "Couldn't open file \"", p_cFilename[0], "\"." );
      throw( stcErr );
    }
    
    //Open the subplot streams
    for (i = 1; i < iNumOutFiles; i++) {
      //Tack on the file extension to the root for the whole-plot file
      sprintf( p_cFilename[i], "%s%s%s%s", m_cFileRoot, "_", mp_subplots[i - 1].cSubplotName, DETAILED_OUTPUT_FILE_EXT );
      p_outFile[i] = fopen( p_cFilename[i], "wt" ); //overwrite existing
      //If we can't open the file throw an error
      if ( NULL == p_outFile[i] )
      {
        modelErr stcErr;
        stcErr.iErrorCode = BAD_FILE;
        strcpy( stcErr.cFunction, "clOutput::WriteDetailedOutputHeader" );
        sprintf( stcErr.cMoreInfo, "%s%s%s", "Couldn't open file \"", p_cFilename[i], "\"." );
        throw( stcErr );
      }
    }

    //Open parameter file
    inFile = fopen( mp_oSimManager->GetParFilename(), "rt" );

    //If we can't open the file throw an error
    if ( NULL == inFile )
    {
      modelErr stcErr;
      stcErr.iErrorCode = BAD_FILE;
      strcpy( stcErr.cFunction, "clOutput::WriteDetailedOutputHeader" );
      sprintf( stcErr.cMoreInfo, "%s%s%s", "Couldn't open file \"", mp_oSimManager->GetParFilename(), "\"." );
      throw( stcErr );
    }

    //Get the file code string
    AssembleFileCode( detailed_output, DETAILED_OUTPUT_FILE_VERSION, cFileCode );

    //Write the detailed output opening tag
    sprintf( cBuf, "%s%s%s", "<?xml version=\"1.0\" encoding=\"UTF-8\"?><rundata fileCode=\"", cFileCode, "\">" );
    for (i = 0; i < iNumOutFiles; i++) {
      fwrite( cBuf, strlen( cBuf ), 1, p_outFile[i] );
    }

    //Now start the data feed from parameter file to detailed output file.  We
    //want to start at the <paramFile> tag.
    bReading = false;
    iNumRead = BUFFER_SIZE - 1;

    while ( iNumRead == BUFFER_SIZE - 1 )
    { //this is our end-of-file test
      iNumRead = fread( cBuf, sizeof( char ), BUFFER_SIZE - 1, inFile );
      cBuf[iNumRead] = '\0';

      if ( !bReading )
      {

        //We haven't found the parameter file opener tag yet - so find it first
        cPos = strstr( cBuf, "<paramFile" );
        if ( cPos != NULL )
        {

          //It's in this string - find the index in the string where cPos is
          iPos = -1;
          iStrLen = strlen( cBuf );
          for ( i = 0; i < iStrLen; i++ )
          {
            if ( cPos == cBuf + i )
            {
              iPos = i;
              break;
            }
          }

          if ( iPos > -1 )
          {
            for (i = 0; i < iNumOutFiles; i++) {
              fwrite( cPos, strlen( cBuf ) - iPos, 1, p_outFile[i] );
            }
            bReading = true;
          }
        }

      }
      else
      {

        //We've already found the parameter string - so it's a straight
        //read/write
        for (i = 0; i < iNumOutFiles; i++) {
          fwrite( cBuf, strlen( cBuf ), 1, p_outFile[i] );
        }
      }
    }


    //Write detailed output closing tag

    strcpy( cBuf, "</rundata>" );
    for (i = 0; i < iNumOutFiles; i++) {
      fwrite( cBuf, strlen( cBuf ), 1, p_outFile[i] );
    }

    fclose( inFile );
    
    for (i = 0; i < iNumOutFiles; i++) {
      fclose( p_outFile[i] );
    }

    //Zip and tar the whole-plot file
    ZipFile( p_cFilename[0], mp_oSimManager->GetAppPath() );
    strcat( p_cFilename[0], GZIP_EXT );
    AddFileToNewTarball( m_cTarball, p_cFilename[0], mp_oSimManager->GetAppPath() );
    DeleteThisFile( p_cFilename[0] ); //it's copied to tarball - so remove original
    
    //Repeat with the subplot files
    for (i = 1; i < iNumOutFiles; i++) {
      ZipFile( p_cFilename[i], mp_oSimManager->GetAppPath() );
      strcat( p_cFilename[i], GZIP_EXT );
      AddFileToNewTarball( mp_subplots[i - 1].cSubplotTarball, p_cFilename[i], mp_oSimManager->GetAppPath() );
      DeleteThisFile( p_cFilename[i] ); //it's copied to tarball - so remove original
    }
    
    delete[] p_outFile;
    for (i = 0; i < iNumOutFiles; i++) {
      delete[] p_cFilename[i];
    }
    delete[] p_cFilename;

  } //end of try block
  catch ( modelErr & err )
  {
    for (i = 0; i < iNumOutFiles; i++) {
      delete[] p_cFilename[i];
    }
    delete[] p_cFilename;
    delete[] p_outFile;
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    for (i = 0; i < iNumOutFiles; i++) {
      delete[] p_cFilename[i];
    }
    delete[] p_cFilename;
    delete[] p_outFile;
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clOutput::WriteDetailedOutputHeader" );
    throw( stcErr );
  }
}


/////////////////////////////////////////////////////////////////////////////
// Action()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
/////////////////////////////////////////////////////////////////////////////
void clOutput::Action()
{
  try
  {
    if ( strlen( m_cFileRoot ) == 0 ) return;

    //Get the filename of the detailed output timestep file
    int iTimestep = mp_oSimManager->GetCurrentTimestep(), 
        iNumOutputFiles = m_iNumSubplotsToSave + 1,
        i;
    char **p_cTimestepFilename;
    
    p_cTimestepFilename = new char*[iNumOutputFiles];
    for (i = 0; i < iNumOutputFiles; i++) {
      p_cTimestepFilename[i] = new char[MAX_FILENAME_SIZE];
      GetTimestepFilename( iTimestep, i - 1, p_cTimestepFilename[i] );
    }

    //Write the header for each file
    for (i = 0; i < iNumOutputFiles; i++) {
      WriteTimestepHeader( p_cTimestepFilename[i] );
    }
    
    //Write the tree data - WriteTreeData will take care of file juggling
    WriteTreeData( p_cTimestepFilename );
    
    //Write grid data to the whole plot file only
    WriteGridData( p_cTimestepFilename[0] );
    
    //Write the footer for each file
    for (i = 0; i < iNumOutputFiles; i++) {
      WriteTimestepFooter( p_cTimestepFilename[i] );
    }
    
    //Zip each file and add it to the tarball
    ZipFile( p_cTimestepFilename[0], mp_oSimManager->GetAppPath() );
    strcat( p_cTimestepFilename[0], GZIP_EXT ); //'cause that's the new filename after
    AddFileToTarball( m_cTarball, p_cTimestepFilename[0], mp_oSimManager->GetAppPath() );
    DeleteThisFile( p_cTimestepFilename[0] );
    delete[] p_cTimestepFilename[0];
    
    //Repeat with the subplot files
    for (i = 1; i < iNumOutputFiles; i++) {
      ZipFile( p_cTimestepFilename[i], mp_oSimManager->GetAppPath() );
      strcat( p_cTimestepFilename[i], GZIP_EXT );
      AddFileToTarball( mp_subplots[i - 1].cSubplotTarball, p_cTimestepFilename[i], mp_oSimManager->GetAppPath() );
      DeleteThisFile( p_cTimestepFilename[i] ); //it's copied to tarball - so remove original
      delete[] p_cTimestepFilename[i];
    }
    
    delete[] p_cTimestepFilename; 

  } //end of try block
  catch ( modelErr & err )
  {
    throw( err );
  } //throw without clearing this object
  catch ( modelMsg & msg )
  {
    throw( msg );
  }
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clOutput::Action" );
    throw( stcErr );
  }
}


/////////////////////////////////////////////////////////////////////////////
// WriteTreeData()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
// October 24, 2005 - Changed the way this searched for trees (LEM)
// May 10, 2006 - Added support for subplots (LEM)
/////////////////////////////////////////////////////////////////////////////
void clOutput::WriteTreeData( char ** p_cFilename )
{
  clTreePopulation * p_oTrees = ( clTreePopulation * ) mp_oSimManager->GetPopulationObject( "treepopulation" ); //pointer to tree population
  clTreeSearch * p_oTreesToWrite; //pointer to search results
  clTree * p_oTree;
  FILE ** p_out = NULL; //out file to write to
  char cTemp[100], //temp for string formatting
       cTempTreeVal[MAX_TREE_CHAR_VALUE_SIZE], //for getting char values
       ** p_cBuf = NULL; //file writing buffer for each output file
  float fTemp, //for getting float values
        fX, fY; //tree coordinates 
  int iTimestep = mp_oSimManager->GetCurrentTimestep(), //current timestep
       iNumTimesteps = mp_oSimManager->GetNumberOfTimesteps(), //total number of timesteps
       iGridCellSize = p_oTrees->GetGridCellSize(),
       iTemp, //for getting integer values
       iX, iY, //tree grid cell
       iNumOutFiles = m_iNumSubplotsToSave + 1,
       iSp, iTp, iBuf, j; //loop counters
  bool bTemp; //for getting bool values
    
  try
  {

    //Open the out file(s)
    if ( strlen( m_cFileRoot ) == 0 ) return;

    p_out = new FILE*[iNumOutFiles];
    p_cBuf = new char*[iNumOutFiles];
    for (iBuf = 0; iBuf < iNumOutFiles; iBuf++) {
      p_out[iBuf] = fopen( p_cFilename[iBuf], "at" );
      //If the file could not be opened - throw error
      if ( NULL == p_out[iBuf] )
      {
        modelErr stcErr;
        stcErr.iErrorCode = BAD_FILE;
        strcpy( stcErr.cFunction, "clOutputCFile::WriteTreeData" );
        sprintf( stcErr.cMoreInfo, "%s%s%s", "Could not open output file \"", p_cFilename[iBuf], "\"" );
        throw( stcErr );
      }
      
      p_cBuf[iBuf] = new char[BUFFER_SIZE + 1];
      //Clear the buffer
      p_cBuf[iBuf][0] = '\0';
    }

    //Figure out which species/types to write this timestep
    bTemp = false;
    for ( iSp = 0; iSp < m_iNumSpecies; iSp++ )
      for ( iTp = 0; iTp < m_iNumTypes; iTp++ )
      {
        if ( mp_treeSettings[iSp] [iTp].iSaveFreq > -1 )
        {
          mp_treeSettings[iSp] [iTp].iSumTimestep++;
          if ( 0 == iTimestep || iNumTimesteps == iTimestep
               || ( mp_treeSettings[iSp] [iTp].iSaveFreq == mp_treeSettings[iSp] [iTp].iSumTimestep ) )
               {
                 mp_treeSettings[iSp] [iTp].bSaveThisTimeStep = true;
                 bTemp = true;
                 //Reset the counter
                 mp_treeSettings[iSp] [iTp].iSumTimestep = 0;
          }
          else
            mp_treeSettings[iSp] [iTp].bSaveThisTimeStep = false;
        }
      }

    //******************************
    // Write the tree map header
    //******************************
    if ( false == bTemp ) {
      for (iBuf = 0; iBuf < iNumOutFiles; iBuf++) {
        fclose( p_out[iBuf] );
      }
      return;
    }
    
    for (iBuf = 0; iBuf < iNumOutFiles; iBuf++) {
      strcpy( p_cBuf[iBuf], "<tr_treemap>" );

      //Write a species list
      strcpy( cTemp, "<tm_speciesList>" );
      AddToBuffer( p_cBuf[iBuf], cTemp, p_out[iBuf], BUFFER_SIZE );
      for ( j = 0; j < m_iNumSpecies; j++ )
      {
        sprintf( cTemp, "%s%s%s", "<tm_species speciesName=\"", p_oTrees->TranslateSpeciesCodeToName( j ), "\"/>" );
        AddToBuffer( p_cBuf[iBuf], cTemp, p_out[iBuf], BUFFER_SIZE );
      }
      strcpy( cTemp, "</tm_speciesList>" );
      AddToBuffer( p_cBuf[iBuf], cTemp, p_out[iBuf], BUFFER_SIZE );
      

      //Tree settings for each species/type combo
      for ( iSp = 0; iSp < m_iNumSpecies; iSp++ ) 
      {
        for ( iTp = 0; iTp < m_iNumTypes; iTp++ )
        {
          if ( mp_treeSettings[iSp] [iTp].iSaveFreq > -1 )
          {
            //Tree settings opening tag
            sprintf( cTemp, "%s%s%s%d%s", "<tm_treeSettings sp=\"", p_oTrees->TranslateSpeciesCodeToName( iSp ),
                 "\" tp=\"", iTp, "\">" );

            AddToBuffer( p_cBuf[iBuf], cTemp, p_out[iBuf], BUFFER_SIZE );


            //Data member labels and codes
            if ( mp_treeSettings[iSp] [iTp].iNumInts > 0 )
            { //ints

              strcpy( cTemp, "<tm_intCodes>" );
              AddToBuffer( p_cBuf[iBuf], cTemp, p_out[iBuf], BUFFER_SIZE );

              for ( j = 0; j < mp_treeSettings[iSp] [iTp].iNumInts; j++ )
              {
                sprintf( cTemp, "%s%s%s%d%s", "<tm_intCode label=\"",
                     p_oTrees->GetIntDataLabel( mp_treeSettings[iSp] [iTp].p_iIntCodes[j], iSp, iTp ),
                     "\">", mp_treeSettings[iSp] [iTp].p_iIntCodes[j], "</tm_intCode>" );

                AddToBuffer( p_cBuf[iBuf], cTemp, p_out[iBuf], BUFFER_SIZE );
              }

              strcpy( cTemp, "</tm_intCodes>" );
              AddToBuffer( p_cBuf[iBuf], cTemp, p_out[iBuf], BUFFER_SIZE );

            } //end of if (mp_treeSettings[i].iNumInts > 0)
            if ( mp_treeSettings[iSp] [iTp].iNumFloats > 0 )
            { //floats

              strcpy( cTemp, "<tm_floatCodes>" );
              AddToBuffer( p_cBuf[iBuf], cTemp, p_out[iBuf], BUFFER_SIZE );

              for ( j = 0; j < mp_treeSettings[iSp] [iTp].iNumFloats; j++ )
              {
                sprintf( cTemp, "%s%s%s%d%s", "<tm_floatCode label=\"",
                     p_oTrees->GetFloatDataLabel( mp_treeSettings[iSp] [iTp].p_iFloatCodes[j], iSp, iTp ),
                     "\">", mp_treeSettings[iSp] [iTp].p_iFloatCodes[j], "</tm_floatCode>" );

                AddToBuffer( p_cBuf[iBuf], cTemp, p_out[iBuf], BUFFER_SIZE );
              }

              strcpy( cTemp, "</tm_floatCodes>" );
              AddToBuffer( p_cBuf[iBuf], cTemp, p_out[iBuf], BUFFER_SIZE );

            } //end of if (mp_treeSettings[iSp] [iTp].iNumFloats > 0)

            if ( mp_treeSettings[iSp] [iTp].iNumChars > 0 )
            { //chars

              strcpy( cTemp, "<tm_charCodes>" );
              AddToBuffer( p_cBuf[iBuf], cTemp, p_out[iBuf], BUFFER_SIZE );

              for ( j = 0; j < mp_treeSettings[iSp] [iTp].iNumChars; j++ )
              {

                sprintf( cTemp, "%s%s%s%d%s", "<tm_charCode label=\"",
                     p_oTrees->GetCharDataLabel( mp_treeSettings[iSp] [iTp].p_iCharCodes[j], iSp, iTp ),
                     "\">", mp_treeSettings[iSp] [iTp].p_iCharCodes[j], "</tm_charCode>" );

                AddToBuffer( p_cBuf[iBuf], cTemp, p_out[iBuf], BUFFER_SIZE );
              }
              strcpy( cTemp, "</tm_charCodes>" );
              AddToBuffer( p_cBuf[iBuf], cTemp, p_out[iBuf], BUFFER_SIZE );
            } //end of if (mp_treeSettings[iSp] [iTp].iNumChars > 0)

            if ( mp_treeSettings[iSp] [iTp].iNumBools > 0 )
            { //bools

              strcpy( cTemp, "<tm_boolCodes>" );
              AddToBuffer( p_cBuf[iBuf], cTemp, p_out[iBuf], BUFFER_SIZE );

              for ( j = 0; j < mp_treeSettings[iSp] [iTp].iNumBools; j++ )
              {
                sprintf( cTemp, "%s%s%s%d%s", "<tm_boolCode label=\"",
                     p_oTrees->GetBoolDataLabel( mp_treeSettings[iSp] [iTp].p_iBoolCodes[j], iSp, iTp ),
                     "\">", mp_treeSettings[iSp] [iTp].p_iBoolCodes[j], "</tm_boolCode>" );

                AddToBuffer( p_cBuf[iBuf], cTemp, p_out[iBuf], BUFFER_SIZE );

              }

              strcpy( cTemp, "</tm_boolCodes>" );
              AddToBuffer( p_cBuf[iBuf], cTemp, p_out[iBuf], BUFFER_SIZE );
            } //end of if (mp_treeSettings[iSp] [iTp].iNumBools > 0)

            //Tree settings closing tag

            strcpy( cTemp, "</tm_treeSettings>" );
            AddToBuffer( p_cBuf[iBuf], cTemp, p_out[iBuf], BUFFER_SIZE );
          }  
        }
      }  
    }
 
    //******************************
    // Tree data saving - go through all trees and write the ones being saved
    //******************************

    //Find the trees
    p_oTreesToWrite = p_oTrees->Find( "all" );

    //Write the trees
    p_oTree = p_oTreesToWrite->NextTree();
    while ( p_oTree )
    {
      iSp = p_oTree->GetSpecies();
      iTp = p_oTree->GetType();
        
      //Get the tree's grid cell counters
      p_oTree->GetValue(p_oTrees->GetXCode(iSp, iTp), &fX);
      p_oTree->GetValue(p_oTrees->GetYCode(iSp, iTp), &fY);
      iX = (int)floor( fX / iGridCellSize );
      iY = (int)floor( fY / iGridCellSize );
        
      if ( mp_treeSettings[iSp] [iTp].bSaveThisTimeStep )
      {
        for (iBuf = 0; iBuf < iNumOutFiles; iBuf++) {
          
          if (0 == iBuf || mp_subplots[iBuf-1].p_bUseCell[iX][iY]) {

            //Write to the whole timestep, and to each subplot of which this 
            //tree is a part
            sprintf( cTemp, "%s%d%s%d%s", "<tree sp=\"", iSp, "\" tp=\"", iTp, "\">" );
            AddToBuffer( p_cBuf[iBuf], cTemp, p_out[iBuf], BUFFER_SIZE );

            //Write int values for this tree
            for ( j = 0; j < mp_treeSettings[iSp] [iTp].iNumInts; j++ )
            {
              p_oTree->GetValue( mp_treeSettings[iSp] [iTp].p_iIntCodes[j], & iTemp );
              sprintf( cTemp, "%s%d%s%d%s", "<int c=\"", mp_treeSettings[iSp] [iTp].p_iIntCodes[j], "\">", iTemp, "</int>" );
              AddToBuffer( p_cBuf[iBuf], cTemp, p_out[iBuf], BUFFER_SIZE );
            }
            //Write float values for this tree
            for ( j = 0; j < mp_treeSettings[iSp] [iTp].iNumFloats; j++ )
            {
              p_oTree->GetValue( mp_treeSettings[iSp] [iTp].p_iFloatCodes[j], & fTemp );

              sprintf( cTemp, "%s%d%s%g%s", "<fl c=\"", mp_treeSettings[iSp] [iTp].p_iFloatCodes[j], "\">", fTemp, "</fl>" );
              AddToBuffer( p_cBuf[iBuf], cTemp, p_out[iBuf], BUFFER_SIZE );
            }
            //Write char values for this tree
            for ( j = 0; j < mp_treeSettings[iSp] [iTp].iNumChars; j++ )
            {
              p_oTree->GetValue( mp_treeSettings[iSp] [iTp].p_iCharCodes[j], cTempTreeVal );
              if ( strlen( cTempTreeVal ) != 0 )
              {
                sprintf( cTemp, "%s%d%s%s%s", "<ch c=\"", mp_treeSettings[iSp] [iTp].p_iCharCodes[j], "\">",
                     cTempTreeVal, "</ch>" );

                AddToBuffer( p_cBuf[iBuf], cTemp, p_out[iBuf], BUFFER_SIZE );
              }
            }
            //Write bool values for this tree
            for ( j = 0; j < mp_treeSettings[iSp] [iTp].iNumBools; j++ )
            {
              p_oTree->GetValue( mp_treeSettings[iSp] [iTp].p_iBoolCodes[j], & bTemp );
              if ( true == bTemp )
                sprintf( cTemp, "%s%d%s", "<bl c=\"", mp_treeSettings[iSp] [iTp].p_iBoolCodes[j], "\">true</bl>" );
              else
                sprintf( cTemp, "%s%d%s", "<bl c=\"", mp_treeSettings[iSp] [iTp].p_iBoolCodes[j], "\">false</bl>" );

              AddToBuffer( p_cBuf[iBuf], cTemp, p_out[iBuf], BUFFER_SIZE );
            }

            strcpy( cTemp, "</tree>" );
            AddToBuffer( p_cBuf[iBuf], cTemp, p_out[iBuf], BUFFER_SIZE );
          }
        }
      }

      p_oTree = p_oTreesToWrite->NextTree();
    }


    //Write the end tag for the tree map
    //Flush all buffers one last time and close them
    for (iBuf = 0; iBuf < iNumOutFiles; iBuf++) {
      fwrite( p_cBuf[iBuf], strlen( p_cBuf[iBuf] ), 1, p_out[iBuf] );
      strcpy( p_cBuf[iBuf], "</tr_treemap>" );
      fwrite( p_cBuf[iBuf], strlen( p_cBuf[iBuf] ), 1, p_out[iBuf] );    
      fclose( p_out[iBuf] );
    }
    
    delete[] p_out;
    for (iBuf = 0; iBuf < iNumOutFiles; iBuf++) {
      delete[] p_cBuf[iBuf]; 
    }
    delete[] p_cBuf;

  } //end of try block
  catch ( modelErr & err )
  {
    delete[] p_out;
    for (iBuf = 0; iBuf < iNumOutFiles; iBuf++) {
      delete[] p_cBuf[iBuf]; 
    }
    delete[] p_cBuf;
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    delete[] p_out;
    for (iBuf = 0; iBuf < iNumOutFiles; iBuf++) {
      delete[] p_cBuf[iBuf]; 
    }
    delete[] p_cBuf;
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clOutput::WriteTreeData" );
    throw( stcErr );
  }
}


/////////////////////////////////////////////////////////////////////////////
// WriteTimestepHeader()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
// May 10, 2006 - Added support for subplots (LEM)
/////////////////////////////////////////////////////////////////////////////
void clOutput::WriteTimestepHeader( const char * cFilename )
{
  try
  {
    FILE * out; //file writing stream
    char cTemp[50], cFileCode[9]; //file code
    int iFileVersion = 1; //for the filecode

    //If the filename is empty return
    if ( strlen( m_cFileRoot ) == 0 ) return;

    //Open the out file
    out = fopen( cFilename, "wt" );

    //If the file could not be opened - throw error
    if ( NULL == out )
    {
      modelErr stcErr;
      stcErr.iErrorCode = BAD_FILE;
      strcpy( stcErr.cFunction, "clOutput::WriteTimestepHeader" );
      sprintf( stcErr.cMoreInfo, "%s%s%s", "Could not open output file \"", cFilename, "\"" );
      throw( stcErr );
    }

    AssembleFileCode( detailed_output_timestep, iFileVersion, cFileCode );
    //***************************
    //Write the header of the file
    //***************************
    //Root element
    strcpy( cTemp, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>" );
    fwrite( cTemp, strlen( cTemp ), 1, out );

    //  << "<!DOCTYPE timestepRundata SYSTEM \"xml\\RundataTimestep.dtd\">\n"
    sprintf( cTemp, "%s%s%s", "<timestepRundata fileCode=\"", cFileCode, "\">" );
    fwrite( cTemp, strlen( cTemp ), 1, out );

    //Timestep
    sprintf( cTemp, "%s%d%s", "<rt_timestep>", mp_oSimManager->GetCurrentTimestep(), "</rt_timestep>" );
    fwrite( cTemp, strlen( cTemp ), 1, out );

    fclose( out );
  } //end of try block
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clOutput::WriteTimestepHeader" );
    throw( stcErr );
  }
}


/////////////////////////////////////////////////////////////////////////////
// WriteTimestepFooter()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
/////////////////////////////////////////////////////////////////////////////
void clOutput::WriteTimestepFooter( const char * cFilename )
{
  try
  {
    char cCloser[] = "</timestepRundata>";
    FILE * out; //file writing stream

    //If the filename is empty return
    if ( strlen( cFilename ) == 0 ) return;

    //Open the out file
    out = fopen( cFilename, "at" );

    //If we can't open the file throw an error
    if ( NULL == out )
    {
      modelErr stcErr;
      stcErr.iErrorCode = BAD_FILE;
      strcpy( stcErr.cFunction, "clOutput::WriteTimestepFooter" );
      sprintf( stcErr.cMoreInfo, "%s%s%s", "Couldn't open file \"", cFilename, "\"." );
      throw( stcErr );
    }

    fwrite( cCloser, strlen( cCloser ), 1, out );
    fclose( out );

  } //end of try block
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clOutput::WriteTimestepFooter" );
    throw( stcErr );
  }
}


/////////////////////////////////////////////////////////////////////////////
// WriteGridData()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
/////////////////////////////////////////////////////////////////////////////
void clOutput::WriteGridData( const char * cFilename )
{
  try
  {
    char cTempTreeVal[MAX_TREE_CHAR_VALUE_SIZE], //for getting char values
         cBuf[BUFFER_SIZE + 1], //buffer for file writing
         cTemp[100]; //for formatting data into buffer
    clPlot * p_oPlot; //plot object - for grid lengths
    clPackage * p_oPackage; //for writing the grid's packages
    FILE * out; //file to write grid file
    float fPlotLenX, fPlotLenY, //plot dimensions
         fTemp; //for getting float values
    int iTimestep, //current timestep
         iNumTimesteps, //total number of timesteps
         iNumXCells, iNumYCells, //number of X and Y cells in the grid
         iNumDataMembers, //number of a kind of data member
         iTemp, //for getting int values
         i, j, iX, iY; //loop counters
    bool bSavePackages, //for whether or not packages are to be saved
         bTemp; //for getting bool values

    //If the filename is empty return
    if ( strlen( cFilename ) == 0 ) return;

    //Open the out file
    out = fopen( cFilename, "at" );

    //If the file could not be opened - throw error
    if ( NULL == out )
    {
      modelErr stcErr;
      stcErr.iErrorCode = BAD_FILE;
      strcpy( stcErr.cFunction, "clOutput::WriteGridData" );
      sprintf( stcErr.cMoreInfo, "%s%s%s", "Could not open output file \"", cFilename, "\"" );
      throw( stcErr );
    }

    //Get the timestep data
    iTimestep = mp_oSimManager->GetCurrentTimestep();
    iNumTimesteps = mp_oSimManager->GetNumberOfTimesteps();

    //Get the plot dimensions
    p_oPlot = mp_oSimManager->GetPlotObject();
    fPlotLenX = p_oPlot->GetXPlotLength();
    fPlotLenY = p_oPlot->GetYPlotLength();

    //Loop through the grids to be written
    for ( i = 0; i < m_iNumGridsToSave; i++ )
    {
      //Clear the buffer
      cBuf[0] = '\0';

      //Is this a saving timestep?
      //Increment the timestep counter and compare to frequency - always
      //write if this is timestep 0 or the last timestep
      mp_gridSettings[i].iSumTimestep++;
      if ( 0 == iTimestep || iNumTimesteps == iTimestep
           || ( mp_gridSettings[i].iSaveFreq == mp_gridSettings[i].iSumTimestep ) )
           {
             //Reset counter
             mp_gridSettings[i].iSumTimestep = 0;

             //Figure out if there are any package data members to save
             if ( mp_gridSettings[i].iNumPackageInts > 0 || mp_gridSettings[i].iNumPackageFloats > 0
                  || mp_gridSettings[i].iNumPackageChars > 0 || mp_gridSettings[i].iNumPackageBools > 0 )
                    bSavePackages = true;
             else
               bSavePackages = false;

             //********************************
             // Grid header
             //********************************

             //Open tag for grid
             sprintf( cTemp, "%s%s%s", "<grid gridName=\"", mp_gridSettings[i].p_oGridPointer->GetName(), "\">" );
             AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );

             //Data labels - write all even if all are not being saved
             //Ints
             iNumDataMembers = mp_gridSettings[i].p_oGridPointer->GetNumberIntDataMembers();
             if ( iNumDataMembers > 0 )
             {
               strcpy( cTemp, "<ma_intCodes>" );
               AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );

               for ( j = 0; j < iNumDataMembers; j++ )
               {
                 sprintf( cTemp, "%s%s%s%d%s", "<ma_intCode label=\"", mp_gridSettings[i].p_oGridPointer->GetIntDataLabel( j ),
                      "\">", j, "</ma_intCode>" );
                 AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
               }

               strcpy( cTemp, "</ma_intCodes>" );
               AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
             } //end of if (iNumDataMembers > 0)

             //floats
             iNumDataMembers = mp_gridSettings[i].p_oGridPointer->GetNumberFloatDataMembers();
             if ( iNumDataMembers > 0 )
             {
               strcpy( cTemp, "<ma_floatCodes>" );
               AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );

               for ( j = 0; j < iNumDataMembers; j++ )
               {
                 sprintf( cTemp, "%s%s%s%d%s", "<ma_floatCode label=\"", mp_gridSettings[i].p_oGridPointer->GetFloatDataLabel( j ),
                      "\">", j, "</ma_floatCode>" );
                 AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
               }

               strcpy( cTemp, "</ma_floatCodes>" );
               AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
             } //end of if (iNumDataMembers > 0)

             //chars
             iNumDataMembers = mp_gridSettings[i].p_oGridPointer->GetNumberCharDataMembers();
             if ( iNumDataMembers > 0 )
             {
               strcpy( cTemp, "<ma_charCodes>" );
               AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );

               for ( j = 0; j < iNumDataMembers; j++ )
               {
                 sprintf( cTemp, "%s%s%s%d%s", "<ma_charCode label=\"", mp_gridSettings[i].p_oGridPointer->GetCharDataLabel( j ),
                      "\">", j, "</ma_charCode>" );
                 AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
               }

               strcpy( cTemp, "</ma_charCodes>" );
               AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
             } //end of if (iNumDataMembers > 0)

             //bools
             iNumDataMembers = mp_gridSettings[i].p_oGridPointer->GetNumberBoolDataMembers();
             if ( iNumDataMembers > 0 )
             {
               strcpy( cTemp, "<ma_boolCodes>" );
               AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );

               for ( j = 0; j < iNumDataMembers; j++ )
               {
                 sprintf( cTemp, "%s%s%s%d%s", "<ma_boolCode label=\"", mp_gridSettings[i].p_oGridPointer->GetBoolDataLabel( j ),
                      "\">", j, "</ma_boolCode>" );
                 AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
               }

               strcpy( cTemp, "</ma_boolCodes>" );
               AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
             } //end of if (iNumDataMembers > 0)

             //Package data labels
             if ( mp_gridSettings[i].p_oGridPointer->GetPackageDataChanged() )
             {
               //Ints
               iNumDataMembers = mp_gridSettings[i].p_oGridPointer->GetNumberIntPackageDataMembers();
               if ( iNumDataMembers > 0 )
               {
                 strcpy( cTemp, "<ma_packageIntCodes>" );
                 AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );

                 for ( j = 0; j < iNumDataMembers; j++ )
                 {
                   sprintf( cTemp, "%s%s%s%d%s", "<ma_intCode label=\"", mp_gridSettings[i].p_oGridPointer->GetPackageIntDataLabel( j ),
                        "\">", j, "</ma_intCode>" );
                   AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
                 }

                 strcpy( cTemp, "</ma_packageIntCodes>" );
                 AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
               } //end of if (iNumDataMembers > 0)

               //Floats
               iNumDataMembers = mp_gridSettings[i].p_oGridPointer->GetNumberFloatPackageDataMembers();
               if ( iNumDataMembers > 0 )
               {
                 strcpy( cTemp, "<ma_packageFloatCodes>" );
                 AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );

                 for ( j = 0; j < iNumDataMembers; j++ )
                 {
                   sprintf( cTemp, "%s%s%s%d%s", "<ma_floatCode label=\"", mp_gridSettings[i].p_oGridPointer->GetPackageFloatDataLabel( j ),
                        "\">", j, "</ma_floatCode>" );
                   AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
                 }

                 strcpy( cTemp, "</ma_packageFloatCodes>" );
                 AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
               } //end of if (iNumDataMembers > 0)

               //Chars
               iNumDataMembers = mp_gridSettings[i].p_oGridPointer->GetNumberCharPackageDataMembers();
               if ( iNumDataMembers > 0 )
               {
                 strcpy( cTemp, "<ma_packageCharCodes>" );
                 AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );

                 for ( j = 0; j < iNumDataMembers; j++ )
                 {
                   sprintf( cTemp, "%s%s%s%d%s", "<ma_charCode label=\"", mp_gridSettings[i].p_oGridPointer->GetPackageCharDataLabel( j ),
                        "\">", j, "</ma_charCode>" );
                   AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
                 }

                 strcpy( cTemp, "</ma_packageCharCodes>" );
                 AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
               } //end of if (iNumDataMembers > 0)

               //Bools
               iNumDataMembers = mp_gridSettings[i].p_oGridPointer->GetNumberBoolPackageDataMembers();
               if ( iNumDataMembers > 0 )
               {
                 strcpy( cTemp, "<ma_packageBoolCodes>" );
                 AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );

                 for ( j = 0; j < iNumDataMembers; j++ )
                 {
                   sprintf( cTemp, "%s%s%s%d%s", "<ma_boolCode label=\"", mp_gridSettings[i].p_oGridPointer->GetPackageBoolDataLabel( j ),
                        "\">", j, "</ma_boolCode>" );
                   AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
                 }

                 strcpy( cTemp, "</ma_packageBoolCodes>" );
                 AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
               } //end of if (iNumDataMembers > 0)
             }

             //Plot size values
             sprintf( cTemp, "%s%g%s%g%s", "<ma_plotLenX>", fPlotLenX, "</ma_plotLenX><ma_plotLenY>",
                  fPlotLenY, "</ma_plotLenY>" );
             AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );

             //Length of grids
             sprintf( cTemp, "%s%g%s%g%s", "<ma_lengthXCells>", mp_gridSettings[i].p_oGridPointer->GetLengthXCells(),
                  "</ma_lengthXCells><ma_lengthYCells>", mp_gridSettings[i].p_oGridPointer->GetLengthYCells(),
                  "</ma_lengthYCells>" );
             AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );

             //********************************
             // Map Values
             //********************************
             //Escape out if there's nothing to save
             if ( mp_gridSettings[i].iNumInts == 0 && mp_gridSettings[i].iNumFloats == 0 && mp_gridSettings[i].iNumChars == 0
                  && mp_gridSettings[i].iNumBools == 0 && mp_gridSettings[i].iNumPackageInts == 0
                  && mp_gridSettings[i].iNumPackageFloats == 0 && mp_gridSettings[i].iNumPackageChars == 0
                  && mp_gridSettings[i].iNumPackageBools == 0 )
                    goto EndTag;

             iNumXCells = mp_gridSettings[i].p_oGridPointer->GetNumberXCells();
             iNumYCells = mp_gridSettings[i].p_oGridPointer->GetNumberYCells();
             for ( iX = 0; iX < iNumXCells; iX++ )
               for ( iY = 0; iY < iNumYCells; iY++ )
               {

                 //opening grid tag
                 sprintf( cTemp, "%s%d%s%d%s", "<ma_v x=\"", iX, "\" y=\"", iY, "\">" );
                 AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );

                 //Write int values for this cell
                 for ( j = 0; j < mp_gridSettings[i].iNumInts; j++ )
                 {

                   mp_gridSettings[i].p_oGridPointer->GetValueOfCell( iX, iY, mp_gridSettings[i].p_iIntCodes[j], & iTemp );

                   sprintf( cTemp, "%s%d%s%d%s", "<int c=\"", mp_gridSettings[i].p_iIntCodes[j], "\">", iTemp, "</int>" );

                   AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
                 }
                 //Write float values for this cell
                 for ( j = 0; j < mp_gridSettings[i].iNumFloats; j++ )
                 {

                   mp_gridSettings[i].p_oGridPointer->GetValueOfCell( iX, iY, mp_gridSettings[i].p_iFloatCodes[j], & fTemp );

                   sprintf( cTemp, "%s%d%s%g%s", "<fl c=\"", mp_gridSettings[i].p_iFloatCodes[j], "\">", fTemp, "</fl>" );

                   AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
                 }
                 //Write char values for this cell - but don't write empty strings
                 //map reading chokes on empty strings
                 for ( j = 0; j < mp_gridSettings[i].iNumChars; j++ )
                 {

                   mp_gridSettings[i].p_oGridPointer->GetValueOfCell( iX, iY,
                        mp_gridSettings[i].p_iCharCodes[j], cTempTreeVal );

                   if ( strlen( cTempTreeVal ) > 0 )
                   {
                     sprintf( cTemp, "%s%d%s%s%s", "<ch c=\"", mp_gridSettings[i].p_iCharCodes[j], "\">",
                          cTempTreeVal, "</ch>" );

                     AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
                   }
                 }
                 //Write bool values for this cell
                 for ( j = 0; j < mp_gridSettings[i].iNumBools; j++ )
                 {

                   mp_gridSettings[i].p_oGridPointer->GetValueOfCell( iX, iY, mp_gridSettings[i].p_iBoolCodes[j], & bTemp );

                   if ( true == bTemp )
                     sprintf( cTemp, "%s%d%s", "<bl c=\"", mp_gridSettings[i].p_iBoolCodes[j], "\">true</bl>" );

                   else
                     sprintf( cTemp, "%s%d%s", "<bl c=\"", mp_gridSettings[i].p_iBoolCodes[j], "\">false</bl>" );

                   AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
                 }

                 //Packages
                 if ( bSavePackages )
                 {
                   p_oPackage = mp_gridSettings[i].p_oGridPointer->GetFirstPackageOfCell( iX, iY );
                   while ( p_oPackage )
                   {

                     strcpy( cTemp, "<pkg>" );
                     AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );

                     //Write int values for this package
                     for ( j = 0; j < mp_gridSettings[i].iNumPackageInts; j++ )
                     {

                       p_oPackage->GetValue( mp_gridSettings[i].p_iPackageIntCodes[j], & iTemp );

                       sprintf( cTemp, "%s%d%s%d%s", "<pint c=\"", mp_gridSettings[i].p_iPackageIntCodes[j],
                            "\">", iTemp, "</pint>" );

                       AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
                     }
                     //Write float values for this package
                     for ( j = 0; j < mp_gridSettings[i].iNumPackageFloats; j++ )
                     {

                       p_oPackage->GetValue( mp_gridSettings[i].p_iPackageFloatCodes[j], & fTemp );

                       sprintf( cTemp, "%s%d%s%g%s", "<pfl c=\"", mp_gridSettings[i].p_iPackageFloatCodes[j],
                            "\">", fTemp, "</pfl>" );

                       AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
                     }
                     //Write char values for this package
                     for ( j = 0; j < mp_gridSettings[i].iNumPackageChars; j++ )
                     {

                       p_oPackage->GetValue( mp_gridSettings[i].p_iPackageCharCodes[j], cTempTreeVal );

                       if ( strlen( cTempTreeVal ) > 0 )
                       {

                         sprintf( cTemp, "%s%d%s%s%s", "<pch c=\"", mp_gridSettings[i].p_iPackageCharCodes[j], "\">",
                              cTempTreeVal, "</pch>" );

                         AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
                       }
                     }
                     //Write bool values for this package
                     for ( j = 0; j < mp_gridSettings[i].iNumPackageBools; j++ )
                     {

                       p_oPackage->GetValue( mp_gridSettings[i].p_iPackageBoolCodes[j], & bTemp );

                       if ( true == bTemp )
                         sprintf( cTemp, "%s%d%s", "<pbl c=\"", mp_gridSettings[i].p_iPackageBoolCodes[j], "\">true</pbl>" );
                       else
                         sprintf( cTemp, "%s%d%s", "<pbl c=\"", mp_gridSettings[i].p_iPackageBoolCodes[j], "\">false</pbl>" );

                       AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
                     }
                     strcpy( cTemp, "</pkg>" );
                     AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );

                     p_oPackage = p_oPackage->GetNextPackage();
                   } //end of while (p_oPackage)
                 } //end of if (bSavePackages)

                 //closing value tag
                 strcpy( cTemp, "</ma_v>" );
                 AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
               }
             //closing grid tag
           EndTag:
             //Flush buffer for the last time
             fwrite( cBuf, strlen( cBuf ), 1, out );
             strcpy( cBuf, "</grid>" );
             fwrite( cBuf, strlen( cBuf ), 1, out );
      } //end of if this is a saving timestep
    } //end of for (i = 0; i < m_iNumGridsToSave; i++)

    fclose( out );
  } //end of try block
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clOutput::WriteGridData" );
    throw( stcErr );
  }
}


/////////////////////////////////////////////////////////////////////////////
// GetTimestepFilename()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
// May 10, 2006 - Added subplot support (LEM)
/////////////////////////////////////////////////////////////////////////////
void clOutput::GetTimestepFilename( int iTimestep, int iSubplot, char * cFilename )
{
  try
  {
    //If the filename is empty return empty string
    if ( strlen( m_cFileRoot ) == 0 )
    {
      //Empty the filename string passed
      cFilename[0] = '\0';
      return;
    }

    if (-1 == iSubplot) {
      sprintf( cFilename, "%s%s%d%s", m_cFileRoot, "_", iTimestep, DETAILED_OUTPUT_FILE_EXT );
    } else if (iSubplot > -1 && iSubplot < m_iNumSubplotsToSave) {
      sprintf( cFilename, "%s%s%s%s%d%s", m_cFileRoot, "_", mp_subplots[iSubplot].cSubplotName, "_", iTimestep, DETAILED_OUTPUT_FILE_EXT ); 
    } else {
      //Error
      modelErr stcErr;
      stcErr.iErrorCode = BAD_DATA;
      sprintf( stcErr.cMoreInfo, "%s%d", "Invalid subplot number:", iSubplot );
      strcpy( stcErr.cFunction, "clOutput::GetTimestepFilename" );
      throw( stcErr ); 
    }

  } //end of try block
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clOutput::GetTimestepFilename" );
    throw( stcErr );
  }
}
