//---------------------------------------------------------------------------
// Storm.cpp
//---------------------------------------------------------------------------
#include "Storm.h"
#include "ParsingFunctions.h"
#include "SimManager.h"
#include "ModelMath.h"
#include "GridBase.h"

//////////////////////////////////////////////////////////////////////////////
// Constructor
// Edit history:
// -------------
// September 9, 2004 - Created (LEM)
//////////////////////////////////////////////////////////////////////////////
clStorm::clStorm( clSimManager * p_oSimManager ) : clWorkerBase( p_oSimManager ), clBehaviorBase( p_oSimManager )
{
  try
  {
    //Set the namestring
    strcpy( m_cNameString, "storm" );

    //Allowed file types
    m_iNumAllowedTypes = 2;
    mp_iAllowedFileTypes = new int[m_iNumAllowedTypes];
    mp_iAllowedFileTypes[0] = parfile;
    mp_iAllowedFileTypes[1] = detailed_output;

    //Versions
    m_fVersionNumber = 1;
    m_fMinimumVersionNumber = 1;

    mp_fStormProbabilities = NULL;
    mp_oStormGrid = NULL;
    mp_oSusceptibilityMap = NULL;

    //Hard-coded values
    m_iNumSeverityClasses = 10;
  }
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clStorm::clStorm" );
    throw( stcErr );
  }
}

//////////////////////////////////////////////////////////////////////////////
// Destructor
// Edit history:
// -------------
// September 9, 2004 - Created (LEM)
//////////////////////////////////////////////////////////////////////////////
clStorm::~clStorm()
{
  delete[] mp_fStormProbabilities;
}

/////////////////////////////////////////////////////////////////////////////
// GetData()
// Edit history:
// -------------
// September 9, 2004 - Created (LEM)
/////////////////////////////////////////////////////////////////////////////

void clStorm::GetData( xercesc::DOMDocument * p_oDoc )
{
  try
  {
    ReadParFile( p_oDoc );
    DoGridSetup();
    CalculateStormProbabilities();
    SetStochFuncPointer();
  }
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clStorm::GetData" );
    throw( stcErr );
  }

}

/////////////////////////////////////////////////////////////////////////////
// ReadParFile()
// Edit history:
// -------------
// September 9, 2004 - Created (LEM)
// November 9, 2005 - Made it so return intervals could equal 0 (LEM)
/////////////////////////////////////////////////////////////////////////////

void clStorm::ReadParFile( xercesc::DOMDocument * p_oDoc )
{
  try
  {
    DOMElement * p_oElement = p_oDoc->getDocumentElement();
    char cLabel[23];
    float fTemp;
    int iTemp;
    int i;

    //Declare the mp_fStormProbabilities array, where we'll put the return
    //intervals until they're transformed into probabilities
    mp_fStormProbabilities = new float[m_iNumSeverityClasses];

    //Get the parameter file values for the return intervals
    for ( i = 0; i < m_iNumSeverityClasses; i++ )
    {
      sprintf( cLabel, "%s%d%s", "st_s", i + 1, "ReturnInterval" );
      FillSingleValue( p_oElement, cLabel, &fTemp, true );
      //Validate that all values are greater than or equal to 0
      if ( fTemp < 0 )
      {
        modelErr stcErr;
        strcpy( stcErr.cFunction, "clStorm::ReadParFile" );
        strcpy( stcErr.cMoreInfo, "Storm return intervals cannot be negative." );
        stcErr.iErrorCode = BAD_DATA;
        throw( stcErr );
      }
      mp_fStormProbabilities[i] = fTemp;
    }

    //Get the susceptibility
    FillSingleValue( p_oElement, "st_susceptibility", & iTemp, true );

    //Make sure the value is valid and assign
    if ( mapped == iTemp )
    {
      m_iSusceptibility = mapped;
    }
    else if (uniform == iTemp )
    {
      m_iSusceptibility = uniform;
    }
    else
    {
      //Unrecognized value - throw error
      modelErr stcErr;
      strcpy( stcErr.cFunction, "clStorm::ReadParFile" );
      sprintf( stcErr.cMoreInfo, "%s%d", "Unrecognized value for susceptibility: ", iTemp );
      stcErr.iErrorCode = BAD_DATA;
      throw( stcErr );
    }

    //Get the stochasticity
    FillSingleValue( p_oElement, "st_stochasticity", & iTemp, true );

    if ( deterministic == iTemp )
    {
      m_iStochasticity = deterministic;
    }
    else if ( stochastic == iTemp )
    {
      m_iStochasticity = stochastic;
    }
    else
    {
      //Unrecognized value - throw error
      modelErr stcErr;
      strcpy( stcErr.cFunction, "clStorm::ReadParFile" );
      sprintf( stcErr.cMoreInfo, "%s%d", "Unrecognized value for stochasticity: ", iTemp );
      stcErr.iErrorCode = BAD_DATA;
      throw( stcErr );
    }

    //If the damage pattern is stochastic, get the probability distribution
    //function
    if ( stochastic == m_iStochasticity )
    {

      FillSingleValue( p_oElement, "st_probFunction", & iTemp, true );

      //Make sure the value is valid and assign
      if ( lognormal == iTemp )
      {
        m_iDistribution = lognormal;
      }
      else if ( normal == iTemp )
      {
        m_iDistribution = normal;
      }
      else if ( negative_binomial == iTemp )
      {
        m_iDistribution = negative_binomial;
      }
      else
      {
        //Unrecognized value - throw error
        modelErr stcErr;
        strcpy( stcErr.cFunction, "clStorm::ReadParFile" );
        sprintf( stcErr.cMoreInfo, "%s%d",
             "Unrecognized value for damage pattern probability distribution function: ", iTemp );
        stcErr.iErrorCode = BAD_DATA;
        throw( stcErr );
      }

      //If the value was normal or lognormal, get the standard deviation;
      //if it is negative binomial, get the clumping parameter
      if ( normal == m_iDistribution || lognormal == m_iDistribution )
      {
        FillSingleValue( p_oElement, "st_standardDeviation", & mp_fStdDev, true );
      }
      else if ( negative_binomial == m_iDistribution )
      {
        FillSingleValue( p_oElement, "st_clumpingParameter", & mp_fClump, true );
      }

    }

  }
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clStorm::ReadParFile" );
    throw( stcErr );
  }
}

/////////////////////////////////////////////////////////////////////////////
// DoGridSetup()
// Edit history:
// -------------
// September 9, 2004 - Created (LEM)
/////////////////////////////////////////////////////////////////////////////

void clStorm::DoGridSetup()
{
  try
  {

    //If the susceptibility is "mapped", then we require the existence of the
    //"Storm Susceptibility" grid.  So get it and throw an error if it doesn't
    //exist.
    if ( mapped == m_iSusceptibility )
    {
      mp_oSusceptibilityMap = mp_oSimManager->GetGridObject( "Storm Susceptibility" );
      if ( NULL == mp_oSusceptibilityMap )
      {
        modelErr stcErr;
        strcpy( stcErr.cFunction, "clStorm::DoGridSetup" );
        strcpy( stcErr.cMoreInfo, "A grid map for the grid \"Storm Susceptibility\" is required for this file." );
        stcErr.iErrorCode = BAD_DATA;
        throw( stcErr );
      }

      //Get the index for the "index" data member and throw an error if it
      //doesn't exist
      m_iSusceptIndexCode = mp_oSusceptibilityMap->GetFloatDataCode( "index" );
      if ( -1 == m_iSusceptIndexCode )
      {
        modelErr stcErr;
        strcpy( stcErr.cFunction, "clStorm::DoGridSetup" );
        strcpy( stcErr.cMoreInfo, "The grid map for the grid \"Storm Susceptibility\" is missing the data member\"index\"." );
        stcErr.iErrorCode = BAD_DATA;
        throw( stcErr );
      }
    }

    //Check to see if the user put in grid data for "Storm Damage", in which
    //case it would already be created (ignore any actual values in this case)
    mp_oStormGrid = mp_oSimManager->GetGridObject( "Storm Damage" );
    if ( NULL != mp_oStormGrid )
    {

      //If there was also a susceptibility grid, make sure that the
      //number of grid cells are the same
      if ( NULL != mp_oSusceptibilityMap )
      {
        if ( mp_oSusceptibilityMap->GetNumberXCells() != mp_oStormGrid->GetNumberXCells()
             || mp_oSusceptibilityMap->GetNumberYCells() != mp_oStormGrid->GetNumberYCells() )
             {
               modelErr stcErr;
               strcpy( stcErr.cFunction, "clStorm::DoGridSetup" );
               strcpy( stcErr.cMoreInfo,
                    "The grid cell resolutions for \"Storm Susceptibility\" and \"Storm Damage\" must match." );
               stcErr.iErrorCode = BAD_DATA;
               throw( stcErr );
        }
      }

      m_iDmgIndexCode = mp_oStormGrid->GetFloatDataCode( "dmg_index" );
      m_iStormTimeCode = mp_oStormGrid->GetFloatDataCode( "stormtime" );
      if ( -1 == m_iDmgIndexCode)
      {
        modelErr stcErr;
        strcpy( stcErr.cFunction, "clStorm::DoGridSetup" );
        strcpy( stcErr.cMoreInfo, "The grid map for the grid \"Storm Damage\" is missing the data member\"dmg_index\"." );
        stcErr.iErrorCode = BAD_DATA;
        throw( stcErr );
      }
      if ( -1 == m_iStormTimeCode)
      {
        modelErr stcErr;
        strcpy( stcErr.cFunction, "clStorm::DoGridSetup" );
        strcpy( stcErr.cMoreInfo, "The grid map for the grid \"Storm Damage\" is missing the data member\"stormtime\"." );
        stcErr.iErrorCode = BAD_DATA;
        throw( stcErr );
      }
    }
    else
    {

      //The grid "Storm Damage" was not created, so create it now.  If the
      //grid "Storm Susceptibility" exists, use its grid resolution; otherwise
      //use the default
      float fXCellLength = 0, fYCellLength = 0;
      if ( NULL != mp_oSusceptibilityMap )
      {
        fXCellLength = mp_oSusceptibilityMap->GetLengthXCells();
        fYCellLength = mp_oSusceptibilityMap->GetLengthYCells();
      }
      mp_oStormGrid = mp_oSimManager->CreateGrid( "Storm Damage", //grid name
           0, //number integers
           2, //number floats
           0, //number chars
           0, //number bools
           fXCellLength, fYCellLength );

      m_iDmgIndexCode = mp_oStormGrid->RegisterFloat( "dmg_index" );
      m_iStormTimeCode = mp_oStormGrid->RegisterFloat( "stormtime" );
    }

    //Initialize values in "Storm Damage"
    TimestepCleanup();

  }
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clStorm::DoGridSetup" );
    throw( stcErr );
  }
}

/////////////////////////////////////////////////////////////////////////////
// TimestepCleanup()
// Edit history:
// -------------
// September 9, 2004 - Created (LEM)
/////////////////////////////////////////////////////////////////////////////
void clStorm::TimestepCleanup()
{
  float fVal = 0;
  int iX, iY, iNumXCells, iNumYCells;
  iNumXCells = mp_oStormGrid->GetNumberXCells();
  iNumYCells = mp_oStormGrid->GetNumberYCells();
  for ( iX = 0; iX < iNumXCells; iX++ )
  {
    for ( iY = 0; iY < iNumYCells; iY++ )
    {
      mp_oStormGrid->SetValueOfCell( iX, iY, m_iDmgIndexCode, fVal );
    }
  }
}

/////////////////////////////////////////////////////////////////////////////
// CalculateStormProbabilities()
// Edit history:
// -------------
// September 9, 2004 - Created (LEM)
/////////////////////////////////////////////////////////////////////////////
void clStorm::CalculateStormProbabilities()
{
  int i;

  //Calculate each probability
  for ( i = 0; i < m_iNumSeverityClasses; i++ )
  {
    if (mp_fStormProbabilities[i] != 0) {
      mp_fStormProbabilities[i] = ( float )1 / mp_fStormProbabilities[i];
    }
  }
}

/////////////////////////////////////////////////////////////////////////////
// Action()
// Edit history:
// -------------
// September 9, 2004 - Created (LEM)
/////////////////////////////////////////////////////////////////////////////
void clStorm::Action()
{
  try
  {
    float fNumberYearsPerTimestep = mp_oSimManager->GetNumberOfYearsPerTimestep();
    int i, j;
    bool bStorms = false;  //whether or not we had a storm this timestep
    for ( i = 0; i < m_iNumSeverityClasses; i++ )
    {
      for ( j = 0; j < fNumberYearsPerTimestep; j++ )
      {
        if ( clModelMath::GetRand() <= mp_fStormProbabilities[i] )
        {
          ApplyDamage( i );
          bStorms = true;
        }
      }
    }
    
    AdjustTimeSinceLastStormCounter(bStorms);
    
  }
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clStorm::Action" );
    throw( stcErr );
  }
}

/////////////////////////////////////////////////////////////////////////////
// SetStochFuncPointer()
// Edit history:
// -------------
// September 9, 2004 - Created (LEM)
/////////////////////////////////////////////////////////////////////////////
void clStorm::SetStochFuncPointer()
{
  if ( stochastic != m_iStochasticity ) return;

  if ( lognormal == m_iDistribution )
  {
    RandomDraw = & clStorm::LognormalDraw;
  }
  else if ( normal == m_iDistribution )
  {
    RandomDraw = & clStorm::NormalDraw;
  }
  else if ( negative_binomial == m_iDistribution )
  {
    RandomDraw = & clStorm::NegBinomialDraw;
  }
}


/////////////////////////////////////////////////////////////////////////////
// ApplyDamage()
// Edit history:
// -------------
// September 9, 2004 - Created (LEM)
/////////////////////////////////////////////////////////////////////////////
void clStorm::ApplyDamage( int iSeverityClass )
{
  float fDamage, fMeanSeverity;
  int iX, iY, iNumXCells, iNumYCells;
  iNumXCells = mp_oStormGrid->GetNumberXCells();
  iNumYCells = mp_oStormGrid->GetNumberYCells();

  //Get the storm's mean severity
  fMeanSeverity = ( ( float )iSeverityClass / 10 ) + (clModelMath::GetRand()/10);

  //Calculate the appropriate damage depending on the damage pattern
  if ( deterministic == m_iStochasticity )
  {
    if ( uniform == m_iSusceptibility )
    {

      //Deterministic uniform - each value in the grid cell gets this storm's
      //damage index.  Remember that storm damages are additive to a max of 1.
      //Since all grids get the same value, get that value for the first cell
      //and apply to all cells
      mp_oStormGrid->GetValueOfCell( 0, 0, m_iDmgIndexCode, & fDamage );
      fDamage += fMeanSeverity;
      if ( fDamage > 1 ) fDamage = 1;

      for ( iX = 0; iX < iNumXCells; iX++ )
      {
        for ( iY = 0; iY < iNumYCells; iY++ )
        {
          mp_oStormGrid->SetValueOfCell( iX, iY, m_iDmgIndexCode, fDamage );
        }
      }
    }
    else if ( mapped == m_iSusceptibility )
    {
      //Deterministic mapped
      float fSusceptibility;
      for ( iX = 0; iX < iNumXCells; iX++ )
      {
        for ( iY = 0; iY < iNumYCells; iY++ )
        {

          //Get the susceptibility index for this cell
          mp_oSusceptibilityMap->GetValueOfCell( iX, iY, m_iSusceptIndexCode, & fSusceptibility );

          //Get the new damage and add it
          mp_oStormGrid->GetValueOfCell( iX, iY, m_iDmgIndexCode, & fDamage );
          fDamage += fSusceptibility * fMeanSeverity;
          if ( fDamage > 1 ) fDamage = 1;
          mp_oStormGrid->SetValueOfCell( iX, iY, m_iDmgIndexCode, fDamage );
        }
      }
    }
  }
  else if ( stochastic == m_iStochasticity )
  {

    if ( uniform == m_iSusceptibility )
    {
      //Stochastic uniform
      float fNewDamage;
      for ( iX = 0; iX < iNumXCells; iX++ )
      {
        for ( iY = 0; iY < iNumYCells; iY++ )
        {

          //Get the new storm's damage value
          fNewDamage = ( * this.*RandomDraw ) ( fMeanSeverity );

          //Add it to the existing cell damage and bound it at 1
          mp_oStormGrid->GetValueOfCell( iX, iY, m_iDmgIndexCode, & fDamage );
          fDamage += fNewDamage;
          if ( fDamage > 1 ) fDamage = 1;

          mp_oStormGrid->SetValueOfCell( iX, iY, m_iDmgIndexCode, fDamage );
        }
      }
    }
    else if ( mapped == m_iSusceptibility )
    {
      //Stochastic mapped
      float fSusceptibility, fNewDamage;
      for ( iX = 0; iX < iNumXCells; iX++ )
      {
        for ( iY = 0; iY < iNumYCells; iY++ )
        {

          //Get the susceptibility index for this cell
          mp_oSusceptibilityMap->GetValueOfCell( iX, iY, m_iSusceptIndexCode, & fSusceptibility );

          //Get the new storm's damage value
          fNewDamage = ( * this.*RandomDraw ) ( fMeanSeverity );

          //Get the new damage and add it
          mp_oStormGrid->GetValueOfCell( iX, iY, m_iDmgIndexCode, & fDamage );
          fDamage += fSusceptibility * fNewDamage;
          if ( fDamage > 1 ) fDamage = 1;
          mp_oStormGrid->SetValueOfCell( iX, iY, m_iDmgIndexCode, fDamage );
        }
      }
    }
  }
}

/////////////////////////////////////////////////////////////////////////////
// AdjustTimeSinceLastStormCounter()
// Edit history:
// -------------
// February 17, 2006 - Created (LEM)
/////////////////////////////////////////////////////////////////////////////
void clStorm::AdjustTimeSinceLastStormCounter(bool bStormThisTimestep)
{
  float fCounter,
        fNumYearsPerTimestep = mp_oSimManager->GetNumberOfYearsPerTimestep();
  int iX, iY, iNumXCells = mp_oStormGrid->GetNumberXCells(),
      iNumYCells = mp_oStormGrid->GetNumberYCells();

  if (bStormThisTimestep) {

    fCounter = 0;
  } 
  else {
  	mp_oStormGrid->GetValueOfCell( 0, 0, m_iStormTimeCode, & fCounter );
  	fCounter += fNumYearsPerTimestep;
  }
  
  for ( iX = 0; iX < iNumXCells; iX++ )
  {
    for ( iY = 0; iY < iNumYCells; iY++ )
    {
      mp_oStormGrid->SetValueOfCell( iX, iY, m_iStormTimeCode, fCounter );
    }
  }
}
